__all__ = ('AuditLog', 'AuditLogEntry', 'AuditLogIterator', 'AuditLogChange', )

import warnings

from ...env import API_VERSION

from ..utils import Unknown, now_as_id, id_to_datetime
from ..core import CHANNELS, USERS, ROLES, MESSAGES, SCHEDULED_EVENTS
from ..permission import Permission
from ..color import Color
from ..user import User, ClientUserBase
from ..webhook import Webhook
from ..permission import PermissionOverwrite
from ..integration import Integration
from ..bases import Icon, maybe_snowflake
from ..channel import VideoQualityMode, ChannelThread
from ..scheduled_event import ScheduledEventStatus, ScheduledEventEntityType, PrivacyLevel
from ..emoji import create_unicode_emoji

from .utils import create_partial_guild_from_id
from .guild import SystemChannelFlag, Guild
from .preinstanced import AuditLogEvent, VerificationLevel, ContentFilterLevel, MessageNotificationLevel, VoiceRegion, \
    MFA


class AuditLog:
    """
    Whenever an admin action is performed on the API, an audit log entry is added to the respective guild's audit
    logs. This class represents a requested  collections of these entries.
    
    Attributes
    ----------
    entries : `list` of ``AuditLogEntry``
        A list of audit log entries, what the audit log contains.
    guild : ``Guild``
        The audit logs' respective guild.
    integrations : `dict` of (`int`, ``Integration``) items
        A dictionary what contains the mentioned integrations by the audit log's entries. The keys are the `id`-s of
        the integrations, meanwhile the values are the integrations themselves.
    threads : `dict` of (`int`, ``ChannelThread``) items
        A dictionary containing the mentioned threads inside of the audit logs.
    users : `dict` of (`int`, ``ClientUserBase``) items
        A dictionary, what contains the mentioned users by the audit log's entries. The keys are the `id`-s of the
        users, meanwhile the values are the users themselves.
    webhooks : `dict` of (`int`, ``Webhook``) items
        A dictionary what contains the mentioned webhook by the audit log's entries. The keys are the `id`-s of the
        webhooks, meanwhile the values are the values themselves.
    """
    __slots__ = ('entries', 'guild', 'integrations', 'threads', 'users', 'webhooks')
    def __init__(self, data, guild):
        """
        Creates an ``AuditLog`` instance from the data received from Discord.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Data received from Discord.
        guild : ``Guild``
            The respective guild of the audit logs.
        """
        self.guild = guild
        
        self.users = users = {}
        try:
            user_datas = data['users']
        except KeyError:
            pass
        else:
            for user_data in user_datas:
                user = User(user_data)
                users[user.id] = user
        
        self.webhooks = webhooks = {}
        try:
            webhook_datas = data['webhook']
        except KeyError:
            pass
        else:
            for webhook_data in webhook_datas:
                webhook = Webhook(webhook_data)
                webhooks[webhook.id] = webhook
        
        self.integrations = integrations = {}
        try:
            integration_datas = data['integrations']
        except KeyError:
            pass
        else:
            for integration_data in integration_datas:
                integration = Integration(integration_data)
                integrations[integration.id] = integration
        
        self.threads = threads = {}
        try:
            thread_datas = data['threads']
        except KeyError:
            pass
        else:
            for thread_data in thread_datas:
                thread = ChannelThread(thread_data, None, guild.id)
                threads[thread.id] = thread
        
        self.entries = entries = []
        try:
            entry_datas = data['audit_log_entries']
        except KeyError:
            pass
        else:
            for entry_data in entry_datas:
                entries.append(AuditLogEntry(entry_data, self))
    
    def __iter__(self):
        """Iterates over the audit log's entries."""
        return iter(self.entries)
    
    def __reversed__(self):
        """Reversed iterator over the audit log's entries."""
        return reversed(self.entries)
    
    def __len__(self):
        """Returns the amount of entries, what the audit lgo contain."""
        return len(self.entries)
    
    def __getitem__(self,index):
        """Returns the specific audit log entry at the given index."""
        return self.entries.__getitem__(index)
    
    def __repr__(self):
        """Returns the representation of the Audit log."""
        return f'<{self.__class__.__name__} of {self.guild.name}, length={len(self.entries)}>'


class AuditLogIterator:
    """
    An async iterator over a guild's audit logs.
    
    Attributes
    ----------
    _data : `dict` of (`str`, `Any`) items
        Data to be sent to Discord when requesting an another audit log chunk. Contains some information, which are not
        stored by any attributes of the audit log iterator, these are the filtering `user` and `event` options.
    _index : `int`
        The next audit log entries index to yield.
    client : ``Client``
        The client, who will execute the api requests.
    entries : `list` of ``AuditLogEntry``
        A list of the already received audit log entries.
    guild : ``Guild``
        The audit log iterator's respective guild.
    integrations : `dict` of (`int`, ``Integration``) items
        A dictionary what contains the mentioned integrations by the audit log's entries. The keys are the `id`-s of
        the integrations, meanwhile the values are the integrations themselves.
    threads : `dict` of (`int`, ``ChannelThread``) items
        A dictionary containing the mentioned threads inside of the audit logs.
    users : `dict` of (`int`, ``ClientUserBase`` items
        A dictionary, what contains the mentioned users by the audit log's entries. The keys are the `id`-s of the
        users, meanwhile the values are the users themselves.
    webhooks : `dict` of (`int`, ``Webhook``) items
        A dictionary what contains the mentioned webhook by the audit log's entries. They keys are the `id`-s of the
        webhooks, meanwhile the values are the values themselves.
    """
    __slots__ = ('_data', '_index', 'client', 'entries', 'guild', 'integrations', 'threads', 'users', 'webhooks')
    
    async def __new__(cls, client, guild, user=None, event=None):
        """
        Creates an audit log iterator with the given parameters.
        
        This method is a coroutine.
        
        Parameters
        ----------
        client : ``Client``
            The client, who will execute the api requests.
        guild : ``Guild`` or `int` instance
            The guild, what's audit logs will be requested.
        user : `None`, ``ClientUserBase`` or `int` instance, Optional
            Whether the audit logs should be filtered only to those, which were created by the given user.
        event : `None`, ``AuditLogEvent`` or `int` instance, Optional
            Whether the audit logs should be filtered only on the given event.
        
        Raises
        ------
        TypeError
            - If `guild` was not given neither as ``Guild``, nor as `int` instance.
            - If `user` was not given neither as `None`, ``ClientUserBase`` nor as `int` instance.
            - If `event` as not not given neither as `None`, ``AuditLogEvent`` nor as `int` instance.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        data = {
            'limit': 100,
            'before': now_as_id(),
        }
        
        if isinstance(guild, Guild):
            guild_id = guild.id
        else:
            guild_id = maybe_snowflake(guild)
            if guild_id is None:
                raise TypeError(f'`guild_or_discovery` can be `{Guild.__name__}` or `int` instance, got '
                    f'{guild.__class__.__name__}.')
            
            guild = None
        
        if (user is not None):
            if isinstance(user, ClientUserBase):
                user_id = user.id
            
            else:
                user_id = maybe_snowflake(user)
                if user_id is None:
                    raise TypeError(f'`user` can be given as `{ClientUserBase.__name__}` or `int` instance, '
                        f'got {user.__class__.__name__}.')
            
            data['user_id'] = user_id
        
        if (event is not None):
            if isinstance(event, AuditLogEvent):
                event_value = event.value
            elif isinstance(event, int):
                event_value = event
            else:
                raise TypeError(f'`event` can be given as `None`, `{AuditLogEvent.__name__}` or `int` instance, got '
                    f'{event.__class__.__name__}.')
            
            data['action_type'] = event_value
        
        if guild is None:
            log_data = await client.http.audit_log_get_chunk(guild_id, data)
            if guild is None:
                guild = create_partial_guild_from_id(guild_id)
        else:
            log_data = None
        
        self = object.__new__(cls)
        self._data = data
        self._index = 0
        self.client = client
        self.guild = guild
        self.entries = []
        self.users = {}
        self.webhooks = {}
        self.integrations = {}
        self.threads = {}
        
        if (log_data is not None):
            self._process_data(log_data)
        
        return self
    
    
    async def load_all(self):
        """
        Loads all not yet loaded audit logs of the audit log iterator's guild.
        
        This method is a coroutine.
        """
        entries = self.entries
        client = self.client
        http = client.http
        data = self._data
        
        while True:
            if entries:
                data['before'] = entries[-1].id
            
            log_data = await http.audit_log_get_chunk(self.guild.id, data)
            
            try:
                self._process_data(log_data)
            except StopAsyncIteration:
                return
            
            if len(entries)%100:
                return
    
    
    def transform(self):
        """
        Converts the audit log iterator to an audit log object.
        
        Returns
        -------
        audit_log : ``AuditLog``
        """
        result = object.__new__(AuditLog)
        result.guild = self.guild
        result.entries = self.entries
        result.users = self.users
        result.webhooks = self.webhooks
        result.integrations = self.integrations
        result.threads = self.threads
        return result
    
    def __aiter__(self):
        """Returns self and resets the `.index`."""
        self._index = 0
        return self
    
    async def __anext__(self):
        """
        Yields the next entry of the audit log iterator.
        
        This method is a coroutine.
        """
        ln = len(self.entries)
        index = self._index
        
        if index < ln:
            self._index += 1
            return self.entries[index]
        
        if index%100:
            raise StopAsyncIteration
        
        data = self._data
        if ln:
            data['before'] = self.entries[ln-1].id
        
        log_data = await self.client.http.audit_log_get_chunk(self.guild.id, data)
        self._process_data(log_data)
        self._index += 1
        return self.entries[index]

    def __repr__(self):
        """Returns the representation of the audit log iterator."""
        return f'<{self.__class__.__name__} of {self.client.full_name} at {self.guild.name}>'
    
    def _process_data(self, data):
        """
        Processes a batch of audit log data received from Discord.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Data received from Discord.
        """
        try:
            entry_datas = data['audit_log_entries']
        except KeyError:
            raise StopAsyncIteration from None
        
        if not entry_datas:
            raise StopAsyncIteration
        
        
        try:
            users_data = data['users']
        except KeyError:
            pass
        else:
            users = self.users
            
            for user_data in users_data:
                user = User(user_data)
                users[user.id] = user
        
        
        try:
            webhooks_data = data['webhook']
        except KeyError:
            pass
        else:
            webhooks = self.webhooks
            
            for webhook_data in webhooks_data:
                webhook = Webhook(webhook_data)
                webhooks[webhook.id] = webhook
        
        
        try:
            integration_datas = data['integrations']
        except KeyError:
            pass
        else:
            integrations = self.integrations
            
            for integration_data in integration_datas:
                integration = Integration(integration_data)
                integrations[integration.id] = integration
        
        
        try:
            thread_datas = data['threads']
        except KeyError:
            pass
        else:
            threads = self.threads
            
            for thread_data in thread_datas:
                thread = ChannelThread(thread_data, None, self.guild.id)
                threads[thread.id] = thread
        
        entries = self.entries
        for entry_data in entry_datas:
            entries.append(AuditLogEntry(entry_data, self))


def convert_detail_days(key, value, all_):
    return 'days', int(value)


def convert_detail_users_removed(key, value, all_):
    return 'users_removed', int(value)


def convert_detail_channel(key, value, all_):
    channel_id = int(value)
    try:
        channel = CHANNELS[channel_id]
    except KeyError:
        channel = Unknown('Channel', channel_id)
    return 'channel', channel


def convert_detail_message(key, value, all_):
    message_id = int(value)
    try:
        message = MESSAGES[message_id]
    except KeyError:
        message = Unknown('Message', message_id)
    
    return 'message', message


def convert_detail_amount(key, value, all_):
    return 'amount', int(value)


def convert_detail_permission_overwrite_target(key, value, all_):
    id_ = int(value)
    try:
        type_name = all_['type']
    except KeyError:
        # broken data
        return None
    
    if type_name == 'member':
        try:
            target = USERS[id_]
        except KeyError:
            target = Unknown('User', id_)
    
    elif type_name == 'role':
        try:
            target = ROLES[id_]
        except KeyError:
            target = Unknown('Role', id_, all_.get('name', ''))
    else:
        # permission_overwrite type can be only member and role, so if it is else,
        # the data is broken again
        return None
    
    return 'target', target

def convert_detail_permission_overwrite_processed(key, value, all_):
    return None

DETAIL_CONVERSIONS = {
    'delete_member_days': convert_detail_days,
    'members_removed': convert_detail_users_removed,
    'channel_id': convert_detail_channel,
    'message_id': convert_detail_message,
    'count': convert_detail_amount,
    'id': convert_detail_permission_overwrite_target,
    'type': convert_detail_permission_overwrite_processed,
    'role_name': convert_detail_permission_overwrite_processed,
}

del convert_detail_days
del convert_detail_users_removed
del convert_detail_channel
del convert_detail_message
del convert_detail_amount
del convert_detail_permission_overwrite_target
del convert_detail_permission_overwrite_processed


def DETAIL_CONVERTER_DEFAULT(key, value, all_):
    return key, value


def convert_guild(entry, parent, target_id):
    return parent.guild


def convert_channel(entry, parent, target_id):
    if target_id is None:
        target = None
    else:
        target_id = int(target_id)
        try:
            target = parent.guild.channels[target_id]
        except KeyError:
            target = Unknown('Channel', target_id)
    
    return target


def convert_user(entry, parent, target_id):
    # target_id can be None for any reason
    if target_id is None:
        target = None
    else:
        target_id = int(target_id)
        try:
            target = parent.users[target_id]
        except KeyError:
            target = Unknown('User', target_id)
    
    return target


def convert_role(entry, parent, target_id):
    if target_id is None:
        target = None
    else:
        target_id = int(target_id)
        try:
            target = parent.guild.roles[target_id]
        except KeyError:
            target = Unknown('Role', target_id)
    
    return target


def convert_invite(entry, parent, target_id):
    #every other data is at #change
    for change in entry.changes:
        if change.attribute_name != 'code':
            continue
        
        if entry.type is AuditLogEvent.invite_delete:
            code = change.before
        else:
            code = change.after
        break
    
    else:
        code = '' # malformed ?
    
    return Unknown('Invite', code)


def convert_webhook(entry, parent, target_id):
    if target_id is None:
        target = None
    else:
        target_id = int(target_id)
        try:
            target = parent.webhooks[target_id]
        except KeyError:
            target = Unknown('Webhook', target_id)
    
    return target


def convert_emoji(entry, parent, target_id):
    if target_id is None:
        target = None
    else:
        target_id = int(target_id)
        try:
            target = parent.guild.emojis[target_id]
        except KeyError:
            target = Unknown('Emoji', target_id)
    
    return target


def convert_message(entry, parent, target_id):
    if target_id is None:
        target = None
    else:
        target_id = int(target_id)
        try:
            target = MESSAGES[target_id]
        except KeyError:
            target = Unknown('Message', target_id)
    
    return target


def convert_integration(entry, parent, target_id):
    if target_id is None:
        target = None
    else:
        target_id = int(target_id)
        try:
            target = parent.integrations[target_id]
        except KeyError:
            target = Unknown('Integration', target_id)
    
    return target


def convert_sticker(entry, parent, target_id):
    if target_id is None:
        target = None
    else:
        target_id = int(target_id)
        try:
            target = parent.guild.stickers[target_id]
        except KeyError:
            target = Unknown('Sticker', target_id)
    
    return target


def convert_scheduled_event(entry, parent, target_id):
    if target_id is None:
        target = None
    else:
        target_id = int(target_id)
        try:
            target = SCHEDULED_EVENTS[target_id]
        except KeyError:
            target = Unknown('ScheduledEvent', target_id)
    
    return target


def convert_thread(entry, parent, target_id):
    if target_id is None:
        target = None
    else:
        target_id = int(target_id)
        try:
            target = parent.threads[target_id]
        except KeyError:
            target = Unknown('ChannelThread', target_id)
    
    return target


CONVERSIONS = {
    0: convert_guild,
    1: convert_channel,
    2: convert_user,
    3: convert_role,
    4: convert_invite,
    5: convert_webhook,
    6: convert_emoji,
    7: convert_message,
    8: convert_integration,
    9: convert_sticker,
    10: convert_scheduled_event,
    11: convert_thread,
}

del convert_guild
del convert_channel
del convert_user
del convert_role
del convert_invite
del convert_webhook
del convert_emoji
del convert_message
del convert_integration
del convert_sticker
del convert_scheduled_event
del convert_thread


class AuditLogEntry:
    """
    Represents an entry of an ``AuditLog``.
    
    Attributes
    ----------
    changes : `list` of ``AuditLogChange``
        The changes of the entry.
    details : `None` or `dict` of (`str`, `Any`) items
        Additional information for a specific action types.
    id : `int`
        The unique identifier number of the entry.
    reason : `None` or `str`
        The reason provided with the logged action.
    target : `None`, ``Guild``,  ``ChannelGuildBase`` instance, ``ClientUserBase``, ``Role``, ``Webhook``,
            ``Emoji``, ``Message``, ``Integration``, ``Sticker``, ``Unknown``
        The target entity of the logged action. If the entity is not found, it will be set as an ``Unknown`` instance.
        It can also happen, that target entity is not provided, then target will be set as `None`.
    type : ``AuditLogEvent``
        The event type of the logged action.
    user : `None`, ``ClientUserBase``
        The user, who executed the logged action. If no user is provided then can be `None` as well.
    """
    __slots__ = ('changes', 'details', 'id', 'reason', 'target', 'type', 'user',)
    
    def __init__(self, data, parent):
        """
        Creates an audit log entry, from entry data sent inside of an ``AuditLog``'s data and from the audit itself.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Data received from Discord.
        parent : ``AuditLog`` or ``AuditLogIterator``
            The parent of the entry, what contains the respective guild, the included users, webhooks and the
            integrations to work with.
        """
        self.id = int(data['id'])
        self.type = AuditLogEvent.get(int(data['action_type']))
        
        options = data.get('options', None)
        if (options is None):
            details = None
        else:
            details = {}
            for key, value in options.items():
                result = DETAIL_CONVERSIONS.get(key, DETAIL_CONVERTER_DEFAULT)(key, value, options)
                if result is None:
                    continue
                
                key, value = result
                details[key] = value
            
            if not details:
                details = None
        
        self.details = details
        
        user_id = data.get('user_id', None)
        if user_id is None:
            user = None
        else:
            user = parent.users.get(int(user_id), None)
        self.user = user
        
        self.reason = data.get('reason', None)
        
        change_datas = data.get('changes', None)
        if (change_datas is None):
            changes = None
        else:
            changes = []
            for change_data in change_datas:
                try:
                    key = change_data['key']
                except KeyError: #malformed?
                    continue
                
                change = TRANSFORMERS.get(key, transform_nothing)(key, change_data)
                if (change is not None):
                    changes.append(change)
            
            if not changes:
                changes = None
        
        self.changes = changes
        
        try:
            conversion = CONVERSIONS[self.type.value//10]
        except KeyError:
            target = None
        else:
            target = conversion(self, parent, data.get('target_id', None))
        self.target = target
    
    @property
    def created_at(self):
        """
        When the audit log entry was created.
        
        Returns
        -------
        created_at : `datetime`
        """
        return id_to_datetime(self.id)
    
    def __repr__(self):
        """Returns the representation of the audit log entry."""
        repr_parts = [
            '<', self.__class__.__name__,
            ' id=', repr(self.id),
            ', type=', self.type.name,
        ]
        
        repr_parts.append(', user=')
        user = self.user
        if user is None:
            user_repr = 'None'
        else:
            user_repr = user.full_name
        repr_parts.append(user_repr)
        
        repr_parts.append(', target=')
        repr_parts.append(repr(self.target))
        
        repr_parts.append(', change count=')
        changes = self.changes
        if changes is None:
            change_amount_repr = '0'
        else:
            change_amount_repr = repr(len(self.changes))
        repr_parts.append(change_amount_repr)
        
        reason = self.reason
        if reason is not None:
            repr_parts.append(', reason=')
            # use repr to escape special inserted characters
            repr_parts.append(repr(reason))
        
        details = self.details
        if details is not None:
            repr_parts.append(', details=')
            repr_parts.append(repr(details))
        
        repr_parts.append('>')
        
        return ''.join(repr_parts)


def transform_nothing(name, data):
    before = data.get('old_value', None)
    after = data.get('new_value', None)
    return AuditLogChange(name, before, after)


def transform_deprecated(name, data):
    return None


def transform_icon(name, data):
    if name == 'splash_hash':
        name = 'invite_splash'
    else:
        name = name[:-5]
    
    before = Icon.from_base16_hash(data.get('old_value', None))
    
    after = Icon.from_base16_hash(data.get('new_value', None))
    
    return AuditLogChange(name, before, after)


def transform_bool__separated(name, data):
    before = data.get('old_value', None)
    after = data.get('new_value', None)
    
    return AuditLogChange('separated', before, after)


def transform_channel(name, data):
    name = name[:-3]
    
    value = data.get('old_value', None)
    if value is None:
        before = None
    else:
        value = int(value)
        try:
            before = CHANNELS[value]
        except KeyError:
            before = Unknown('Channel', value)
    
    value = data.get('new_value', None)
    if value is None:
        after = None
    else:
        value = int(value)
        try:
            after = CHANNELS[value]
        except KeyError:
            after = Unknown('Channel', value)
    
    return AuditLogChange(name, before, after)


def transform_color(name, data):
    before = data.get('old_value', None)
    if (before is not None):
        before = Color(before)
    
    after = data.get('new_value', None)
    if (after is not None):
        after = Color(after)
    
    return AuditLogChange('color', before, after)


def transform_content_filter(name, data):
    before = data.get('old_value', None)
    if (before is not None):
        before = ContentFilterLevel.get(before)
    
    after = data.get('new_value', None)
    if (after is not None):
        after = ContentFilterLevel.get(after)
    
    return AuditLogChange('content_filter', before, after)


def transform_video_quality_mode(name, data):
    before = data.get('old_value', None)
    if (before is not None):
        before = VideoQualityMode.get(before)
    
    after = data.get('new_value', None)
    if (after is not None):
        after = VideoQualityMode.get(after)
    
    return AuditLogChange('video_quality_mode', before, after)


def transform_int__days(name, data):
    before = data.get('old_value', None)
    after = data.get('new_value', None)
    return AuditLogChange('days', before, after)


def transform_int__slowmode(name, data):
    before = data.get('old_value', None)
    after = data.get('new_value', None)
    return AuditLogChange('slowmode', before, after)


def transform_message_notification(name, data):
    before = data.get('old_value', None)
    if (before is not None):
        before = MessageNotificationLevel.get(before)
    
    after = data.get('new_value', None)
    if (after is not None):
        after = MessageNotificationLevel.get(after)
    
    return AuditLogChange('message_notification', before, after)


def transform_mfa(name, data):
    before = data.get('old_value', None)
    if (before is not None):
        before = MFA.get(before)
    
    after = data.get('new_value', None)
    if (after is not None):
        after = MFA.get(after)
    
    return AuditLogChange('mfa', before, after)


def transform_overwrites(name, data):
    before = data.get('old_value', None)
    if (before is not None):
        before = [PermissionOverwrite(overwrite_data) for overwrite_data in before]
    
    after = data.get('new_value', None)
    if (after is not None):
        after = [PermissionOverwrite(overwrite_data) for overwrite_data in after]
    
    return AuditLogChange('overwrites', before, after)


def transform_permission(name, data):
    if name.endswith('_new'):
        name = name[:-4]
    
    before = data.get('old_value', None)
    if (before is not None):
        before = Permission(before)
    
    after = data.get('new_value', None)
    if (after is not None):
        after = Permission(after)
    
    return AuditLogChange(name, before, after)


def transform_region(name, data):
    before = data.get('old_value', None)
    if (before is not None):
        before = VoiceRegion.get(before)
    
    after = data.get('new_value', None)
    if (after is not None):
        after = VoiceRegion.get(after)
    
    return AuditLogChange('region', before, after)


def transform_roles(name, data):
    roles = []
    if name == '$add':
        before = None
        after = roles
    else:
        before = roles
        after = None
    
    for element in data['new_value']:
        role_id = int(element['id'])
        try:
            role = ROLES[role_id]
        except KeyError:
            role = Unknown('Role', role_id, element['name'])
        roles.append(role)
    
    return AuditLogChange('role', before, after)


def transform_snowflake(name, data):
    before = data.get('old_value', None)
    if (before is not None):
        before = int(before)
    
    after = data.get('new_value', None)
    if (after is not None):
        after = int(after)
    
    return AuditLogChange(name, before, after)

def transform_str__vanity_code(name, data):
    before = data.get('old_value', None)
    after = data.get('new_value', None)
    return AuditLogChange('vanity_code', before, after)


def transform_system_channel_flags(name, data):
    before = data.get('old_value', None)
    if (before is not None):
        before = SystemChannelFlag(before)
    
    after = data.get('new_value', None)
    if (after is not None):
        after = SystemChannelFlag(after)
    
    return AuditLogChange('system_channel_flags', before, after)


def transform_type(name, data):
    # If we talk about permission overwrite, type can be `str` too, what we ignore
    before = data.get('old_value', None)
    if (before is not None) and isinstance(before, str):
        return None
    
    after = data.get('new_value', None)
    if (after is not None) and isinstance(after,  str):
        return None
    
    return AuditLogChange('type', before, after)


def transform_user(name, data):
    name = name[:-3]
    
    value = data.get('old_value', None)
    if value is None:
        before = None
    else:
        value = int(value)
        try:
            before = USERS[value]
        except KeyError:
            before = Unknown('User', value)
    
    value = data.get('new_value', None)
    if value is None:
        after = None
    else:
        value = int(value)
        try:
            after = USERS[value]
        except KeyError:
            after = Unknown('User', value)
    
    return AuditLogChange(name, before, after)


def transform_verification_level(name, data):
    before = data.get('old_value', None)
    if (before is not None):
        before = VerificationLevel.get(before)
    
    after = data.get('new_value', None)
    if (after is not None):
        after = VerificationLevel.get(after)
    
    return AuditLogChange('verification_level', before, after)


def transform_tags(name, data):
    before = data.get('old_value', None)
    if (before is None) or (not before):
        before = None
    else:
        before = frozenset(before.split(', '))
    
    after = data.get('new_value', None)
    if (after is None) or (not after):
        after = None
    else:
        after = frozenset(after.split(', '))
    
    return AuditLogChange('tags', before, after)


def transform_int__auto_archive_after(name, data):
    before = data.get('old_value', None)
    if (before is not None):
        before *= 60
    
    after = data.get('new_value', None)
    if (after is not None):
        after *= 60
    
    return AuditLogChange('auto_archive_after', before, after)


def transform_int__default_auto_archive_after(name, data):
    before = data.get('old_value', None)
    if (before is not None):
        before *= 60
    
    after = data.get('new_value', None)
    if (after is not None):
        after *= 60
    
    return AuditLogChange('default_auto_archive_after', before, after)


def transform_privacy_level(name, data):
    before = data.get('old_value', None)
    if (before is not None):
        before = PrivacyLevel.get(before)
    
    after = data.get('new_value', None)
    if (after is not None):
        after = PrivacyLevel.get(after)
    
    return AuditLogChange('privacy_level', before, after)


def transform_scheduled_event_status(name, data):
    before = data.get('old_value', None)
    if (before is not None):
        before = ScheduledEventStatus.get(before)
    
    after = data.get('new_value', None)
    if (after is not None):
        after = ScheduledEventStatus.get(after)
    
    return AuditLogChange('status', before, after)


def transform_scheduled_event_entity_type(name, data):
    before = data.get('old_value', None)
    if (before is not None):
        before = ScheduledEventEntityType.get(before)
    
    after = data.get('new_value', None)
    if (after is not None):
        after =  ScheduledEventEntityType.get(after)
    
    return AuditLogChange('entity_type', before, after)


def transform_snowflake_array(name, data):
    before = data.get('old_value', None)
    if (before is None) or (not before):
        before = None
    else:
        before = tuple(int(sub_value) for sub_value in before)
    
    after = data.get('new_value', None)
    if (after is None) or (not after):
        after = None
    else:
        after = tuple(int(sub_value) for sub_value in after)
    
    return AuditLogChange(name, before, after)


def transform_unicode_emoji(name, data):
    before = data.get('old_value', None)
    if (before is None):
        before = None
    else:
        before = create_unicode_emoji(before)
    
    after = data.get('new_value', None)
    if (after is None):
        after = None
    else:
        after = create_unicode_emoji(after)
    
    return AuditLogChange(name, before, after)


TRANSFORMERS = {
    '$add': transform_roles,
    '$remove': transform_roles,
    'account_id': transform_snowflake,
    'afk_channel_id': transform_snowflake,
    'allow': transform_deprecated if API_VERSION in (6, 7) else transform_permission,
    'allow_new': transform_permission if API_VERSION in (6, 7) else transform_deprecated,
    'application_id': transform_snowflake,
    # archived (bool)
    'auto_archive_duration': transform_int__auto_archive_after,
    'avatar_hash': transform_icon,
    'banner_hash': transform_icon,
    # bitrate (int)
    'channel_id': transform_channel,
    # code (str)
    'color': transform_color,
    # deaf (bool)
    # description (None or str)
    'default_message_notifications': transform_message_notification,
    'default_auto_archive_duration': transform_int__default_auto_archive_after,
    'deny': transform_deprecated if API_VERSION in (6, 7) else transform_permission,
    'deny_new': transform_permission if API_VERSION in (6, 7) else transform_deprecated,
    'discovery_splash_hash' : transform_icon,
    # enable_emoticons (bool)
    'entity_type': transform_scheduled_event_entity_type,
    # expire_behavior (int)
    # expire_grace_period (int)
    'explicit_content_filter':transform_content_filter,
    'hoist': transform_bool__separated,
    'icon_hash': transform_icon,
    'id': transform_snowflake,
    'inviter_id': transform_user,
    # locked (bool)
    # mentionable (bool)
    # max_age (int)
    # max_uses (int)
    'mfa_level': transform_mfa,
    # mute (mute)
    # name (str)
    # nick (None or str)
    # nsfw (bool)
    'owner_id': transform_user,
    # position (int)
    'privacy_level': transform_privacy_level,
    'prune_delete_days': transform_int__days,
    'permission_overwrites' : transform_overwrites,
    'permissions': transform_deprecated if API_VERSION in (6, 7) else transform_permission,
    'permissions_new': transform_permission if API_VERSION in (6, 7) else transform_deprecated,
    'public_updates_channel_id' : transform_snowflake,
    'rate_limit_per_user': transform_int__slowmode,
    'region': transform_region,
    'rules_channel_id': transform_snowflake,
    'sku_ids': transform_snowflake_array,
    'splash_hash': transform_icon,
    'status': transform_scheduled_event_status,
    'system_channel_id': transform_snowflake,
    'system_channel_flags': transform_system_channel_flags,
    'tags': transform_tags,
    # temporary (bool)
    # topic (str)
    'type': transform_type,
    'unicode_emoji': transform_unicode_emoji,
    # uses (int)
    'vanity_url_code': transform_str__vanity_code,
    'verification_level': transform_verification_level,
    'video_quality_mode': transform_video_quality_mode,
    'widget_channel_id': transform_snowflake,
    # widget_enabled (bool)
}

del transform_deprecated
del transform_icon
del transform_bool__separated
del transform_channel
del transform_color
del transform_content_filter
del transform_int__days
del transform_int__slowmode
del transform_message_notification
del transform_mfa
del transform_overwrites
del transform_permission
del transform_region
del transform_roles
del transform_snowflake
del transform_str__vanity_code
del transform_system_channel_flags
del transform_type
del transform_user
del transform_verification_level
del transform_video_quality_mode
del transform_tags
del transform_int__auto_archive_after
del transform_int__default_auto_archive_after
del transform_privacy_level
del transform_scheduled_event_status
del transform_scheduled_event_entity_type
del transform_snowflake_array
del transform_unicode_emoji


class AuditLogChange:
    """
    A change of an ``AuditLogEntry``.
    
    Attributes
    ----------
    after : `Any`
        The changed attribute's new value. Defaults to `None`.
    attribute_name : `str`
        The name of the changed attribute.
    before : `Any`
        The changed attribute's original value. Defaults to `None`.
    
    Notes
    -----
    The value of `before` and `after` depending on the value of `attr`. These are:
    
    +-------------------------------+-----------------------------------------------+
    | attribute_name                | before / after                                |
    +===============================+===============================================+
    | account_id                    | `None` or `int`                               |
    +-------------------------------+-----------------------------------------------+
    | afk_channel_id                | `int`                                         |
    +-------------------------------+-----------------------------------------------+
    | allow                         | `None` or ``Permission``                      |
    +-------------------------------+-----------------------------------------------+
    | application_id                | `None` or `int`                               |
    +-------------------------------+-----------------------------------------------+
    | archived                      | `None` or `bool`                              |
    +-------------------------------+-----------------------------------------------+
    | auto_archive_duration         | `None` or `int`                               |
    +-------------------------------+-----------------------------------------------+
    | avatar                        | `None` or ``Icon``                            |
    +-------------------------------+-----------------------------------------------+
    | banner                        | `None` or ``Icon``                            |
    +-------------------------------+-----------------------------------------------+
    | bitrate                       | `None` or `int`                               |
    +-------------------------------+-----------------------------------------------+
    | channel                       | `None` or ``ChannelGuildBase`` instance       |
    +-------------------------------+-----------------------------------------------+
    | code                          | `None` or `str`                               |
    +-------------------------------+-----------------------------------------------+
    | color                         | `None` or ``Color``                           |
    +-------------------------------+-----------------------------------------------+
    | content_filter                | `None` or ``ContentFilterLevel``              |
    +-------------------------------+-----------------------------------------------+
    | days                          | `None` or `int`                               |
    +-------------------------------+-----------------------------------------------+
    | deaf                          | `None` or `bool`                              |
    +-------------------------------+-----------------------------------------------+
    | description                   | `None` or `str`                               |
    +-------------------------------+-----------------------------------------------+
    | default_message_notifications | `None` or ``MessageNotificationLevel``        |
    +-------------------------------+-----------------------------------------------+
    | deny                          | `None` or ``Permission``                      |
    +-------------------------------+-----------------------------------------------+
    | discovery_splash              | `None` or ``Icon``                            |
    +-------------------------------+-----------------------------------------------+
    | enable_emoticons              | `None` or `bool`                              |
    +-------------------------------+-----------------------------------------------+
    | entity_type                   | `None` or ``ScheduledEventEntityType``        |
    +-------------------------------+-----------------------------------------------+
    | expire_behavior               | `None` or `int`                               |
    +-------------------------------+-----------------------------------------------+
    | expire_grace_period           | `None` or `int`                               |
    +-------------------------------+-----------------------------------------------+
    | icon                          | `None` or ``Icon``                            |
    +-------------------------------+-----------------------------------------------+
    | id                            | `None` or `int`                               |
    +-------------------------------+-----------------------------------------------+
    | invite_splash                 | `None` or ``Icon``                            |
    +-------------------------------+-----------------------------------------------+
    | inviter                       | `None`, ``ClientUserBase``                    |
    +-------------------------------+-----------------------------------------------+
    | mentionable                   | `None` or `bool`                              |
    +-------------------------------+-----------------------------------------------+
    | max_age                       | `None` or `int`                               |
    +-------------------------------+-----------------------------------------------+
    | max_uses                      | `None` or `int`                               |
    +-------------------------------+-----------------------------------------------+
    | message_notification          | `None` or ``MessageNotificationLevel``        |
    +-------------------------------+-----------------------------------------------+
    | mfa                           | `None` or ``MFA``                             |
    +-------------------------------+-----------------------------------------------+
    | mute                          | `None` or `bool`                              |
    +-------------------------------+-----------------------------------------------+
    | name                          | `None` or `str`                               |
    +-------------------------------+-----------------------------------------------+
    | nick                          | `None` or `str`                               |
    +-------------------------------+-----------------------------------------------+
    | nsfw                          | `None` or `bool`                              |
    +-------------------------------+-----------------------------------------------+
    | owner                         | `None`, ``ClientUserBase``                    |
    +-------------------------------+-----------------------------------------------+
    | position                      | `None` or `int`                               |
    +-------------------------------+-----------------------------------------------+
    | public_updates_channel_id     | `int`                                         |
    +-------------------------------+-----------------------------------------------+
    | overwrites                    | `None` or `list` of ``PermissionOverwrite``   |
    +-------------------------------+-----------------------------------------------+
    | permissions                   | `None` or ``Permission``                      |
    +-------------------------------+-----------------------------------------------+
    | privacy_level                 | `None` or ``PrivacyLevel``                    |
    +-------------------------------+-----------------------------------------------+
    | region                        | `None` or ``VoiceRegion``                     |
    +-------------------------------+-----------------------------------------------+
    | role                          | `None` or `list` of ``Role``                  |
    +-------------------------------+-----------------------------------------------+
    | rules_channel_id              | `int`                                         |
    +-------------------------------+-----------------------------------------------+
    | separated                     | `None` or `bool`                              |
    +-------------------------------+-----------------------------------------------+
    | slowmode                      | `None` or `int`                               |
    +-------------------------------+-----------------------------------------------+
    | sku_ids                       | `None` or `tuple` of `int`                    |
    +-------------------------------+-----------------------------------------------+
    | system_channel_id             | `int`                                         |
    +-------------------------------+-----------------------------------------------+
    | system_channel_flags          | `None` or ``SystemChannelFlag``               |
    +-------------------------------+-----------------------------------------------+
    | tags                          | `None` ot `frozenset` of `str`                |
    +-------------------------------+-----------------------------------------------+
    | temporary                     | `None` or `bool`                              |
    +-------------------------------+-----------------------------------------------+
    | topic                         | `None` or `str`                               |
    +-------------------------------+-----------------------------------------------+
    | type                          | `None` or `int`                               |
    +-------------------------------+-----------------------------------------------+
    | unicode_emoji                 | `None` or ``Emoji``                           |
    +-------------------------------+-----------------------------------------------+
    | uses                          | `None` or `int`                               |
    +-------------------------------+-----------------------------------------------+
    | vanity_code                   | `None` or `str`                               |
    +-------------------------------+-----------------------------------------------+
    | verification_level            | `None` or ``VerificationLevel``               |
    +-------------------------------+-----------------------------------------------+
    | widget_channel_id             | `int`                                         |
    +-------------------------------+-----------------------------------------------+
    | widget_enabled                | `None` or `bool`                              |
    +-------------------------------+-----------------------------------------------+
    """
    __slots__ = ('after', 'attribute_name', 'before', )
    
    def __init__(self, attribute_name, before, after):
        """
        Creates a new audit log change instance.
        
        Parameters
        ----------
        attribute_name : `str`
            The name of the changed attribute.
        after : `Any`
            The changed attribute's new value.
        before : `Any`
            The changed attribute's original value.
        """
        self.attribute_name = attribute_name
        self.before = before
        self.after = after
    
    def __repr__(self):
        """Returns the representation of the audit log change."""
        return (
            f'{self.__class__.__name__}('
                f'attribute_name={self.attribute_name!r}, '
                f'before={self.before!r}, '
                f'after={self.after!r}'
            f')'
        )
    
    @property
    def attr(self):
        warnings.warn(
            f'`{self.__class__.__name__}.attr` attribute is deprecated, and will be removed in 2021 November. '
            f'Please use `.attribute_name` instead.',
            FutureWarning)
        
        return self.attribute_name



__all__ = ('EmbedCore', )

from ...backend.utils import copy_docs

from ..utils import timestamp_to_datetime, sanitize_mentions, datetime_to_timestamp
from ..color import Color

from .embed_base import EmbedBase, EmbedFooter, EmbedImage, EmbedThumbnail, EmbedVideo, EmbedProvider, EmbedAuthor, \
    EmbedField

class EmbedCore(EmbedBase):
    """
    Represents Discord embedded content. There are two defined embed classes, the other one is ``Embed``.
    
    Each embed what is received from Discord is stored as ``EmbedCore`` object for better operation support. This
    embed type is a valid embed type to send, but it is more cumbersome to build up, because it requires extra
    imports and it is slower to serialize.
    
    Attributes
    ----------
    author : `None` or ``EmbedAuthor``
        Author information.
    color : `None`, ``Color`` or `int`
        The color code of the embed. Passing `0` means black, not like at the case of roles.
    description : `None` or `str`
        The main content of the embed.
    fields : `list` of ``EmbedField``
        Fields' information.
    footer : `None` or ``EmbedFooter``
        Footer information.
    image : `None` or ``EmbedImage``
        Image information.
    provider : `None` or ``EmbedProvider``
        Provider information.
    thumbnail : `None` or ``EmbedThumbnail``
        Thumbnail information.
    timestamp : `None` or `datetime`
        Timestamp of the embed's content. Shows up next to the ``.footer`` separated with a `'|'` character.
    title : `None` or `str`
        The title of the embed. Shows at the top with intense white characters.
    type : `None` or `str`
        The type of the embed. Can be one of `EXTRA_EMBED_TYPES`'s elements. Webhook embeds' type must be `'rich'`.
    url : `None` or `str`
        Url of the embed. If defined, the embed's `title` will show up as a hyper link pointing to the `url`.
    video : `None` or `EmbedVideo`
        Video information.
    """
    __slots__ = ('author', 'color', 'description', 'fields', 'footer', 'image', 'provider', 'thumbnail', 'timestamp',
        'title', 'type', 'url', 'video',)
    
    def __init__(self, title=None, description=None, color=None, url=None, timestamp=None, type_='rich'):
        """
        Creates an embed core instance. Accepts the base parameters of the embed and sets the other ones as `None`.
        
        Parameters
        ----------
        title : `str`, Optional
            The title of the embed. Shows at the top with intense white characters.
        description : `str`, Optional
            The main content of the embed.
        color : ``Color`` or `int`, Optional
            The color code of the embed. Passing `0` means black, not like at the case of roles.
        url : `str`, Optional
            Url of the embed. If defined, the embed's `title` will show up as a hyper link pointing to the `url`.
        timestamp : `datetime`, optional
            Timestamp of the embed's content. Shows up next to the `footer` separated with a `'|'` character.
        type_ : `None` or `str`, Optional
            The type of the embed. Defaults to `'rich'`.
        """
        self.title = title
        self.description = description
        self.color = color
        self.url = url
        self.timestamp = timestamp
        self.type = type_ # must be `rich` for webhook embeds
        self.footer = None
        self.image = None
        self.thumbnail = None
        self.video = None
        self.provider = None
        self.author = None
        self.fields = []
    
    
    @classmethod
    def from_data(cls, data):
        """
        Creates an `EmbedCore`` object from the data sent by Discord.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Embed data received from Discord.
        
        Returns
        -------
        self : ``EmbedCore``
        """
        self = cls.__new__(cls)
        
        self.title = data.get('title', None)
        self.type = data.get('type', None)
        self.description = data.get('description', None)
        self.url = data.get('url', None)

        try:
            timestamp_data = data['timestamp']
        except KeyError:
            timestamp = None
        else:
           timestamp = timestamp_to_datetime(timestamp_data)
        self.timestamp = timestamp
        
        try:
            color_data = data['color']
        except KeyError:
            color = None
        else:
            color = Color(color_data)
        self.color = color

        try:
            footer_data = data['footer']
        except KeyError:
            footer = None
        else:
            footer = EmbedFooter.from_data(footer_data)
        self.footer = footer
        
        try:
            image_data = data['image']
        except KeyError:
            image = None
        else:
            image = EmbedImage.from_data(image_data)
        self.image = image
        
        try:
            thumbnail_data = data['thumbnail']
        except KeyError:
            thumbnail = None
        else:
            thumbnail = EmbedThumbnail.from_data(thumbnail_data)
        self.thumbnail = thumbnail
        
        try:
            video_data = data['video']
        except KeyError:
            video = None
        else:
            video = EmbedVideo.from_data(video_data)
        self.video = video
        
        try:
            provider_data = data['provider']
        except KeyError:
            provider = None
        else:
            provider = EmbedProvider.from_data(provider_data)
        self.provider = provider
        
        try:
            author_data = data['author']
        except KeyError:
            author = None
        else:
            author = EmbedAuthor.from_data(author_data)
        self.author = author
        
        try:
            field_datas = data['fields']
        except KeyError:
            fields = []
        else:
            fields = [EmbedField.from_data(field_data) for field_data in field_datas]
        self.fields = fields
        
        return self
    
    
    def to_data(self):
        """
        Converts the embed core to json serializable `dict` representing it.
        
        Returns
        -------
        data : `dict` of (`str`, `Any`) items
        """
        data = {}
        
        type_ = self.type
        if (type_ is not None):
            data['type'] = type_
        
        title = self.title
        if (title is not None):
            data['title'] = title
        
        description = self.description
        if (description is not None):
            data['description'] = description
            
        color = self.color
        if (color is not None):
            data['color'] = color
        
        url = self.url
        if (url is not None):
            data['url'] = url
        
        timestamp = self.timestamp
        if (timestamp is not None):
            data['timestamp'] = datetime_to_timestamp(timestamp)
        
        footer = self.footer
        if (footer is not None):
            data['footer'] = footer.to_data()
        
        image = self.image
        if (image is not None):
            data['image'] = image.to_data()
        
        thumbnail = self.thumbnail
        if (thumbnail is not None):
            data['thumbnail'] = thumbnail.to_data()
        
        author = self.author
        if (author is not None):
            data['author'] = author.to_data()
        
        fields = self.fields
        if fields:
            data['fields'] = [field.to_data() for field in fields]
        
        return data
    
    
    @copy_docs(EmbedBase.clear)
    def clear(self):
        self.author = None
        self.color = None
        self.description = None
        fields = self.fields
        if (fields is not None):
            fields.clear()
        self.footer = None
        self.image = None
        self.provider = None
        self.thumbnail = None
        self.timestamp = None
        self.title = None
        self.type = None
        self.url = None
        self.video = None
    
    
    @property
    def contents(self):
        """
        Returns the embed's contents.
        
        The embed's contents are the following:
        - `.title`
        - `.description`
        - `.author.name`
        - `.footer.text`
        - `.fields[n].name`
        - `.fields[n].value`
        
        Returns
        -------
        contents : `list` of `str`
        """
        contents = []
        
        title = self.title
        if (title is not None):
            contents.append(title)
        
        description = self.description
        if (description is not None):
            contents.append(description)
        
        author = self.author
        if (author is not None):
            name = author.name
            if (name is not None):
                contents.append(name)
        
        footer = self.footer
        if (footer is not None):
            contents.append(footer.text)
        
        for field in self.fields:
            contents.append(field.name)
            contents.append(field.value)
        
        return contents
    
    
    def _update_sizes(self, data):
        """
        Updates the size information of the embed.
        
        Called when a ``Message`` is edited, but no `edited` timestamp is included with the data. Returns `0` if
        received data does not contain images, if does, then `1`.
        
        This method tries to update the embed's `image`, `thumbnail` amd `video` with their sizes. If any of those is
        not set already (for any reason), then it also creates them.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Embed data received from Discord.
        
        Returns
        -------
        changed : `int`
        """
        changed = 0
        try:
            image_data = data['image']
        except KeyError:
            pass
        else:
            image = self.image
            if image is None:
                self.image = EmbedImage.from_data(image_data)
            else:
                image.height = image_data.get('height', 0)
                image.width = image_data.get('width', 0)
            changed = 1
        
        try:
            thumbnail_data = data['thumbnail']
        except KeyError:
            pass
        else:
            thumbnail = self.thumbnail
            if thumbnail is None:
                self.thumbnail = EmbedThumbnail.from_data(thumbnail_data)
            else:
                thumbnail.height = thumbnail_data.get('height', 0)
                thumbnail.width = thumbnail_data.get('width', 0)
            changed = 1

        try:
            video_data = data['video']
        except KeyError:
            pass
        else:
            video = self.video
            if video is None:
                self.video = EmbedVideo.from_data(video_data)
            else:
                video.height = video_data.get('height', 0)
                video.width = video_data.get('width', 0)
            changed = 1
        
        return changed
    
    
    def _set_sizes(self, data):
        """
        Updates the size information of the embed.
        
        Familiar to ``._update_sizes`` but it not checks whether the embed's images change, just updates them.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Embed data received from Discord.
        """
        try:
            image_data = data['image']
        except KeyError:
            pass
        else:
            image = self.image
            if image is None:
                self.image = EmbedImage.from_data(image_data)
            else:
                image.height = image_data.get('height', 0)
                image.width = image_data.get('width', 0)
        
        try:
            thumbnail_data = data['thumbnail']
        except KeyError:
            pass
        else:
            thumbnail = self.thumbnail
            if thumbnail is None:
                self.thumbnail = EmbedThumbnail.from_data(thumbnail_data)
            else:
                thumbnail.height = thumbnail_data.get('height', 0)
                thumbnail.width = thumbnail_data.get('width', 0)
        
        try:
            video_data = data['video']
        except KeyError:
            pass
        else:
            video = self.video
            if video is None:
                self.video = EmbedVideo.from_data(video_data)
            else:
                video.height = video_data.get('height', 0)
                video.width = video_data.get('width', 0)
    
    
    def _clean_copy(self, message):
        """
        Creates a clean copy of the embed by removing the mentions in it's contents.
        
        Called by ``Message.clean_embeds``.
        
        Parameters
        ----------
        message : ``Message``
            The embed's respective message.
        
        Returns
        -------
        embed : ``EmbedCore``
        """
        new = object.__new__(type(self))
        
        new.title = self.title
        description = self.description
        new.description = None if (description is None) else sanitize_mentions(description, message.guild)
        new.color = self.color
        new.url = self.url
        new.timestamp = self.timestamp
        new.type = self.type
        
        new.footer = self.footer
        new.image = self.image
        new.thumbnail = self.thumbnail
        new.video = self.video
        new.provider = self.provider
        new.author = self.author
        new.fields = [
            type(field)(field.name, sanitize_mentions(field. value, message.guild), inline=field.inline) \
                for field in self.fields]
        
        return new
    
    
    @copy_docs(EmbedBase.copy)
    def copy(self):
        new = object.__new__(type(self))
        
        new.author = self.author
        new.color = self.color
        new.description = self.description
        
        fields = self.fields
        if (fields is not None):
            fields = [field.copy() for field in fields]
        new.fields = fields
        
        new.footer = self.footer
        new.image = self.image
        new.provider = self.provider
        new.thumbnail = self.thumbnail
        new.timestamp = self.timestamp
        new.title = self.title
        new.type = self.type
        new.url = self.url
        new.video = self.video
        
        return new
    
    
    @copy_docs(EmbedBase.copy_with)
    def copy_with(self, **kwargs):
        try:
            author = kwargs.pop('author')
        except KeyError:
            author = self.author
            if (author is not None):
                author = author.copy()
        
        try:
            color = kwargs.pop('color')
        except KeyError:
            color = self.color
        
        try:
            description = kwargs.pop('description')
        except KeyError:
            description = self.description
        
        try:
            fields = kwargs.pop('fields')
        except KeyError:
            fields = self.fields
            if (fields is not None):
                fields = [field.copy() for field in fields]
        
        try:
            footer = kwargs.pop('footer')
        except KeyError:
            footer = self.footer
            if (footer is not None):
                footer = footer.copy()
        
        try:
            image = kwargs.pop('image')
        except KeyError:
            image = self.image
            if (image is not None):
                image = image.copy()
        
        try:
            provider = kwargs.pop('provider')
        except KeyError:
            provider = self.provider
            if (provider is not None):
                provider = provider.copy()
        
        try:
            thumbnail = kwargs.pop('thumbnail')
        except KeyError:
            thumbnail = self.thumbnail
            if (thumbnail is not None):
                thumbnail = thumbnail.copy()
        
        try:
            timestamp = kwargs.pop('timestamp')
        except KeyError:
            timestamp = self.timestamp
        
        try:
            title = kwargs.pop('title')
        except KeyError:
            title = self.title
        
        try:
            type_ = kwargs.pop('type')
        except KeyError:
            type_ = self.type
        
        try:
            url = kwargs.pop('url')
        except KeyError:
            url = self.url
        
        try:
            video = kwargs.pop('video')
        except KeyError:
            video = self.video
            if (video is not None):
                video = video.copy()
        
        if kwargs:
            raise TypeError(f'Unused or unsettable attributes: `{kwargs}`')
        
        new = object.__new__(type(self))
        
        new.author = author
        new.color = color
        new.description = description
        new.fields = fields
        new.footer = footer
        new.image = image
        new.provider = provider
        new.thumbnail = thumbnail
        new.timestamp = timestamp
        new.title = title
        new.type = type_
        new.url = url
        new.video = video
        
        return new



__all__ = ()

from ...backend.utils import KeepType
from ...discord.client import Client

from .extension import EXTENSIONS, EXTENSION_STATE_LOADED

@KeepType(Client)
class Client:

    @property
    def extensions(self):
        """
        Returns a list of extensions added to the client. Added by the `extension_loader` extension.
        
        Returns
        -------
        extensions : `list` of ``Extension``
        """
        extensions = []
        for extension in EXTENSIONS.values():
            if extension._state == EXTENSION_STATE_LOADED:
                snapshot_difference = extension._snapshot_difference
                if (snapshot_difference is not None):
                    for client, client_snapshot_difference in snapshot_difference:
                        if (self is client) and client_snapshot_difference:
                            extensions.append(extension)
                            break
        
        return extensions



__all__ = ('Menu',)

from collections import OrderedDict
from types import MemberDescriptorType

from ....backend.analyzer import CallableAnalyzer
from ....backend.utils import copy_docs
from ....backend.futures import Task, CancelledError

from ....discord.core import KOKORO
from ....discord.interaction.components import _debug_component_components, _debug_component_custom_id, \
    _debug_component_emoji, _debug_component_label, _debug_component_enabled, _debug_component_url, \
    _debug_component_description, _debug_component_default, _debug_component_options, _debug_component_placeholder, \
    _debug_component_min_values, _debug_component_max_values
from ....discord.interaction import ComponentBase, ComponentRow, ComponentButton, ComponentSelect, InteractionEvent, \
    ComponentSelectOption, ComponentType, InteractionType
from ....discord.message import Message
from ....discord.channel import ChannelTextBase
from ....discord.embed import EmbedBase
from ....discord.allowed_mentions import AllowedMentionProxy
from ....discord.client import Client
from ....discord.exceptions import DiscordException, ERROR_CODES

from ..waiters import get_client_from_message, get_client_from_interaction_event, Timeouter

GUI_STATE_NONE = 0
GUI_STATE_READY = 1
GUI_STATE_EDITING = 2
GUI_STATE_CANCELLING = 3
GUI_STATE_CANCELLED = 4
GUI_STATE_SWITCHING_CONTEXT = 5

INTERACTION_TYPE_APPLICATION_COMMAND = InteractionType.application_command
INTERACTION_TYPE_MESSAGE_COMPONENT = InteractionType.message_component

class ComponentSourceIdentityHasher:
    """
    Hasher for components based on their identity.
    
    Attributes
    ----------
    component : ``ComponentBase``
        The stored component by the hasher.
    """
    __slots__ = ('component', )
    def __init__(self, component):
        """
        Creates a new ``ComponentSourceIdentityHasher`` instance with the given parameters.
        
        Parameters
        ----------
        component : ``Component``
            The source component.
        """
        self.component = component
    
    def __hash__(self):
        """Returns the component source identity hasher's hash value."""
        return object.__hash__(self.component)
    
    def __eq__(self, other):
        """Returns whether the two component identity hashers are the same."""
        if type(self) is not type(other):
            return NotImplemented
        
        return self.component is other.component


class ComponentAttributeDescriptor:
    """
    Descriptor for component attribute access.
    
    Attributes
    ----------
    _debugger : `callable`
        Debugger to check whether the given value is correct.
    _is_collection : `None` or ``ComponentBase``
        Sub-component type if applicable.
    _supported_types : `tuple` of `type`
        The supported types of the respective component.
    _name : `str`
        The attribute's name.
    """
    __slots__ = ('_debugger', '_is_collection', '_supported_types', '_name')
    def __new__(cls, name, supported_types, debugger, is_collection):
        """
        Creates a new ``ComponentAttributeDescriptor`` instance with the given parameters.
        
        Parameters
        ----------
        name : `str`
            The attribute's name.
        supported_types : `tuple` of `type`
            The supported component types.
        debugger : `None`, `Function`
            Debugger function to check the passed values in debug mode.
        is_collection : `bool`
            Whether the component is collection of components.
        """
        self = object.__new__(cls)
        self._debugger = debugger
        self._supported_types = supported_types
        self._name = name
        self._is_collection = is_collection
        return self
    
    def __get__(self, instance, type_):
        """
        Returns the component's attribute's value.
        
        Parameters
        ----------
        instance : ``ComponentProxy``
            The parent component proxy.
        type_ : `type` = ``ComponentProxy``
            The component proxy type.
        
        Returns
        -------
        self / attribute_value : ``ComponentAttributeDescriptor`` / `Any`
            Self called from class, or the represented component's attribute's value.
        """
        if instance is None:
            return self
        
        is_collection = self._is_collection
        if is_collection:
            component = instance._get_component()
        else:
            component = instance._get_component_overwrite()
        
        if __debug__:
            if (not isinstance(component, self._supported_types)):
                raise AssertionError(f'{self._name} is not supported for {component.__class__.__name__}.')
        
        attribute_value = getattr(component, self._name)
        if is_collection and (attribute_value is not None):
            menu_instance = instance._instance
            
            attribute_value = tuple(components_descriptor.get_component_proxy(menu_instance) \
                for components_descriptor in attribute_value)
        
        return attribute_value
    
    
    def __set__(self, instance, value):
        """
        Sets the value to the component.
        
        Parameters
        ----------
        instance : ``ComponentProxy``
            The parent component proxy.
        value : `Any`
            The value to set as attribute.
        """
        if __debug__:
            self._debugger(value)
        
        component = instance._get_component_overwrite()
        
        if __debug__:
            if (not isinstance(component, self._supported_types)):
                raise AssertionError(f'{self._name} is not supported for {component.__class__.__name__}.')
        
        
        setattr(component, self._name, value)
    
    def __repr__(self):
        """Returns the component descriptor attribute's representation."""
        repr_parts = [
            '<',
            self.__class__.__name__,
            ' name=',
            repr(self._name),
            ', supported_types=',
            repr(self._supported_types)
        ]
        
        if self._is_collection:
            repr_parts.append(' (collection)')
        
        repr_parts.append('>')
        
        return ''.join(repr_parts)


class ComponentDescriptorState(ComponentBase):
    """
    Stateless component descriptor.
    
    Attributes
    ----------
    _component : ``ComponentBase``
        The wrapped component instance.
    _source_component : ``ComponentBase``
        The source component from which the descriptor is created from.
    _sub_components : `None` or `list` of ``ComponentDescriptor``
        Sub components applicable to the component.
    
    Class Attributes
    ----------------
    custom_id : `NoneType` = `None`
        Placeholder for sub-classes without `custom_id` attribute.
    """
    __slots__ = ('_component', '_source_component', '_sub_components',)
    
    def __new__(cls, source, sub_components):
        """
        Creates a new ``ComponentDescriptor`` instance from the given component.
        
        Parameters
        ----------
        source : ``ComponentBase`` instance
            The source component to create the descriptor from.
        sub_components : `None` or `list` of ``ComponentDescriptor``
            A list of sub components added
        """
        component = source.copy()
        
        self = object.__new__(cls)
        self._component = component
        self._source_component = source
        self._sub_components = sub_components
        
        return self
    
    def get_component_proxy(self, instance):
        """
        Gets component proxy of the descriptor.
        
        Parameters
        ----------
        instance : ``Menu``
            The parent menu instance.
        
        Returns
        -------
        component_proxy : ``ComponentProxy``
        """
        return ComponentProxy(self, instance)

    def __repr__(self):
        """Returns the component descriptor state's representation."""
        repr_parts = [
            '<',
            self.__class__.__name__,
            ', component=', repr(self._component),
        ]
        
        sub_components = self._sub_components
        if (sub_components is not None):
            repr_parts.append(' sub_components=')
            repr_parts.append(repr(sub_components))
        
        repr_parts.append('>')
        
        return ''.join(repr_parts)
    
    
    @property
    def type(self):
        return self._component.type
    
    
    def __getattr__(self, attribute_name):
        """Tries to find the attributes of the respective component."""
        component = self._component
        try:
            getter = type(component).__dict__[attribute_name]
        except KeyError:
            pass
        else:
            if isinstance(getter, MemberDescriptorType):
                return MemberDescriptorType.__get__(getter, component)
        
        raise AttributeError(attribute_name)
    
    
    @copy_docs(ComponentBase.to_data)
    def to_data(self):
        return self._component.to_data()
    
    
    @copy_docs(ComponentBase.__eq__)
    def __eq__(self, other):
        if type(other) is not type(self):
            return NotImplemented
        
        if self._source_component is other._source_component:
            return True
        
        return False
    
    
    @copy_docs(ComponentBase.__hash__)
    def __hash__(self):
        return hash(self._component)
    
    
    @copy_docs(ComponentBase._iter_components)
    def _iter_components(self):
        yield self
        
        sub_components = self._sub_components
        if (sub_components is not None):
            for sub_component in sub_components:
                yield from sub_component._iter_components()
    
    
    @copy_docs(ComponentBase._iter_direct_sub_components)
    def _iter_direct_sub_components(self):
        sub_components = self._sub_components
        if (sub_components is not None):
            yield from sub_components


class ComponentDescriptor(ComponentDescriptorState):
    """
    Descriptor to proxy class attribute component access for each instance.
    
    Attributes
    ----------
    _component : ``ComponentBase``
        The wrapped component instance.
    _source_component : ``ComponentBase``
        The source component from which the descriptor is created from.
    _sub_components : `None` or `list` of ``ComponentDescriptor``
        Sub components applicable to the component.
    _identifier : `int`
        The descriptor's identifier.
    
    Class Attributes
    ----------------
    _identifier_counter : `int`
        Identifier counter for caching.
    """
    _identifier_counter = 0
    
    
    __slots__ = ('_identifier',)
    def __new__(cls, source, sub_components):
        """
        Creates a new ``ComponentDescriptor`` instance from the given component.
        
        Parameters
        ----------
        source : ``ComponentBase`` instance
            The source component to create the descriptor from.
        sub_components : `list` of ``ComponentDescriptor``
            A list of sub components added
        """
        identifier = cls._identifier_counter+1
        cls._identifier_counter = identifier
        
        self = ComponentDescriptorState.__new__(cls, source, sub_components)
        self._identifier = identifier
        return self
    
    def __get__(self, instance, type_):
        """
        Gets the descriptor itself if called from class or a component proxy.
        
        Parameters
        ----------
        instance : ``Menu``
            The menu instance.
        type_ : ``MenuType``
            The menu's type.
        
        Returns
        -------
        self / component_proxy : ``ComponentDescriptor`` / ``ComponentProxy``
            Self if called from class, or the a component proxy for the represented component.
        """
        if instance is None:
            return self
        
        return self.get_component_proxy(instance)
    
    @copy_docs(ComponentDescriptorState.get_component_proxy)
    def get_component_proxy(self, instance):
        component_proxy_cache = instance._component_proxy_cache
        try:
            component_proxy = component_proxy_cache[self._identifier]
        except KeyError:
            component_proxy = component_proxy_cache[self._identifier] = ComponentProxy(self, instance)
        
        return component_proxy
    
    def __repr__(self):
        """Returns the component descriptor's representation."""
        repr_parts = [
            '<',
            self.__class__.__name__,
            ' identifier=', repr(self._identifier),
            ', component=', repr(self._component),
        ]
        
        sub_components = self._sub_components
        if (sub_components is not None):
            repr_parts.append(' sub_components=')
            repr_parts.append(repr(sub_components))
        
        repr_parts.append('>')
        
        return ''.join(repr_parts)


class ComponentProxy(ComponentBase):
    """
    Proxy class for components.
    
    Attributes
    ----------
    _component_overwrite : `None` or ``Component``
        Default component's overwrite.
    _descriptor : ``ComponentDescriptor``
        The creator descriptor, which describes the component's default values.
    _instance : ``Any``
        An instance's of the descriptor's owner type.
    """
    __slots__ = ('_component_overwrite', '_descriptor', '_instance')
    
    def __new__(cls, descriptor, instance):
        """
        Creates a new component proxy.
        
        Parameters
        ----------
        descriptor : ``ComponentDescriptor``
            The creator descriptor, which describes the component's default values.
        instance : ``Any``
            An instance's of the descriptor's owner type.
        """
        self = object.__new__(cls)
        self._descriptor = descriptor
        self._instance = instance
        self._component_overwrite = None
        return self
    
    
    def _iter_component_proxies(self):
        """
        Iterates over the components and sub-component proxies.
        
        This method is an iterable generator.
        
        Yields
        ------
        component_proxy : ``ComponentProxy``
        """
        yield self
        sub_components = self._descriptor._sub_components
        if (sub_components is not None):
            instance = self._instance
            for sub_component in sub_components:
                yield sub_component.get_component_proxy(instance)
    
    
    def _get_component(self):
        """
        Gets the actually used component by the proxy.
        
        Returns
        -------
        component : ``ComponentBase``
        """
        component = self._component_overwrite
        if (component is None):
            component = self._descriptor._component
        
        return component
    
    
    def _get_component_overwrite(self):
        """
        Gets the proxy's component overwrite.
        
        Returns
        -------
        component : ``ComponentBase``
        """
        component = self._component_overwrite
        if (component is None):
            self._component_overwrite = component = self._descriptor._component.copy()
        
        return component
    
    
    @property
    def type(self):
        """Returns the component's type."""
        return self._descriptor._component.type
    
    
    def __repr__(self):
        """Returns the component proxy's representation."""
        component = self._component_overwrite
        if (component is None):
            component = self._descriptor._component
        
        return f'<{self.__class__.__name__} component={component!r}>'
    
    
    @copy_docs(ComponentBase.__hash__)
    def __hash__(self):
        component = self._component_overwrite
        if (component is None):
            component = self._descriptor._component
        
        return hash(component)
    
    
    @copy_docs(ComponentBase.to_data)
    def to_data(self):
        # We will brute force some things. It ain't pretty.
        component_type = type(self._descriptor._component)
        
        if issubclass(component_type, ComponentRow):
            data = {
                'type': ComponentType.row.value,
            }
            
            attribute_value = self.components
            if attribute_value is None:
                field_value = []
            else:
                field_value = [sub_component.to_data() for sub_component in attribute_value]
            data['components'] = field_value
        
        elif issubclass(component_type, ComponentButton):
            data = self._get_component().to_data()
        
        elif issubclass(component_type,  ComponentSelectOption):
            data = self._get_component().to_data()
        
        elif issubclass(component_type,  ComponentSelect):
            data = {
                'type': ComponentType.select.value,
            }
            
            attribute_value = self.options
            if attribute_value is None:
                field_value = []
            else:
                field_value = [sub_component.to_data() for sub_component in attribute_value]
            data['options'] = field_value
            
            component = self._get_component()
            
            data['custom_id'] = component.custom_id
            
            placeholder = component.placeholder
            if (placeholder is not None):
                data['placeholder'] = placeholder
            
            min_values = component.min_values
            if min_values != 1:
                data['min_values'] = min_values
            
            max_values = component.max_values
            if max_values != 1:
                data['max_values'] = max_values
            
            if (not component.enabled):
                data['disabled'] = True
                
        else:
            data = {}
        
        return data
    
    
    components = ComponentAttributeDescriptor(
        'components',
        (ComponentRow,),
        _debug_component_components,
        True,
    )
    
    custom_id = ComponentAttributeDescriptor(
        'custom_id',
        (ComponentButton, ComponentSelect,),
        _debug_component_custom_id,
        False,
    )
    
    emoji = ComponentAttributeDescriptor(
        'emoji',
        (ComponentButton, ComponentSelectOption),
        _debug_component_emoji,
        False,
    )
    
    label = ComponentAttributeDescriptor(
        'label',
        (ComponentButton, ComponentSelectOption),
        _debug_component_label,
        False,
    )

    enabled = ComponentAttributeDescriptor(
        'enabled',
        (ComponentButton, ComponentSelect),
        _debug_component_enabled,
        False,
    )

    url = ComponentAttributeDescriptor(
        'url',
        (ComponentButton,),
        _debug_component_url,
        False,
    )
    
    description = ComponentAttributeDescriptor(
        'description',
        (ComponentSelectOption,),
        _debug_component_description,
        False,
    )
    
    default = ComponentAttributeDescriptor(
        'default',
        (ComponentSelectOption,),
        _debug_component_default,
        False,
    )
    
    options = ComponentAttributeDescriptor(
        'options',
        (ComponentSelect,),
        _debug_component_options,
        True,
    )
    
    placeholder = ComponentAttributeDescriptor(
        'placeholder',
        (ComponentSelect,),
        _debug_component_placeholder,
        False,
    )
    
    min_values = ComponentAttributeDescriptor(
        'min_values',
        (ComponentSelect,),
        _debug_component_min_values,
        False,
    )
    
    max_values = ComponentAttributeDescriptor(
        'max_values',
        (ComponentSelect,),
        _debug_component_max_values,
        False,
    )


def validate_check(check):
    """
    Validates the given check.
    
    Parameters
    ----------
    check : `None` of `callable`
        The check to validate.
    
    Raises
    ------
    TypeError
        If `check` is not `None` neither a non-async function accepting 1 parameter.
    """
    if check is None:
        return
    
    analyzer = CallableAnalyzer(check)
    if analyzer.is_async():
        raise TypeError('`check` should have NOT be be `async` function.')
    
    min_, max_ = analyzer.get_non_reserved_positional_parameter_range()
    if min_ > 2:
        raise TypeError(f'`check` should accept `2` parameters, meanwhile the given callable expects at '
            f'least `{min_!r}`, got `{check!r}`.')
    
    if min_ != 2:
        if max_ < 2:
            if not analyzer.accepts_args():
                raise TypeError(f'`check` should accept `2` parameters, meanwhile the given callable expects '
                    f'up to `{max_!r}`, got `{check!r}`.')


def validate_invoke(invoke):
    """
    Validates the given invoker.
    
    Parameters
    ----------
    invoke : `None` or `async-callable`
        The invoker to validate.
    
    Raises
    ------
    TypeError
        If `invoke` is not `None` nor `async-callable` accepting `2` parameter.
    """
    if invoke is None:
        return
    
    analyzer = CallableAnalyzer(invoke)
    if not analyzer.is_async():
        raise TypeError('`invoke` should have be `async` function.')
    
    min_, max_ = analyzer.get_non_reserved_positional_parameter_range()
    
    if min_ > 2:
        raise TypeError(f'`invoke` should accept `2` parameters, meanwhile the given callable expects at '
            f'least `{min_!r}`, got `{invoke!r}`.')
    
    if min_ != 2:
        if max_ < 2:
            if not analyzer.accepts_args():
                raise TypeError(f'`invoke` should accept `2` parameters, meanwhile the given callable expects '
                    f'up to `{max_!r}`, got `{invoke!r}`.')


def validate_initial_invoke(initial_invoke):
    """
    Validates the given default content getter.
    
    Parameters
    ----------
    initial_invoke : `None` or `async-callable`
        The default content getter to validate.
    
    Raises
    ------
    TypeError
        If `initial_invoke` is not `None` nor `async-callable` accepting `1` parameters.
    """
    if initial_invoke is None:
        return
    
    analyzer = CallableAnalyzer(initial_invoke)
    if not analyzer.is_async():
        raise TypeError('`initial_invoke` should have be `async` function.')
    
    min_, max_ = analyzer.get_non_reserved_positional_parameter_range()
    if min_ > 1:
        raise TypeError(f'`initial_invoke` should accept `1` parameters, meanwhile the given callable expects at '
            f'least `{min_!r}`, got `{initial_invoke!r}`.')
    
    if min_ != 1:
        if max_ < 1:
            if not analyzer.accepts_args():
                raise TypeError(f'`initial_invoke` should accept `1` parameters, meanwhile the given callable '
                    f'expects up to `{max_!r}`, got `{initial_invoke!r}`.')


def validate_get_timeout(get_timeout):
    """
    Validates the given timeout getter.
    
    Parameters
    ----------
    get_timeout : `None` or `callable`
        Timeout getter.
    
    Raises
    ------
    TypeError
        If `get_timeout` is not `None` nor `callable` accepting `1` parameters.
    """
    if get_timeout is None:
        return
    
    analyzer = CallableAnalyzer(get_timeout)
    if analyzer.is_async():
        raise TypeError('`get_timeout` should have be not be async function.')
    
    min_, max_ = analyzer.get_non_reserved_positional_parameter_range()
    if min_ > 1:
        raise TypeError(f'`get_timeout` should accept `1` parameters, meanwhile the given callable expects at '
            f'least `{min_!r}`, got `{get_timeout!r}`.')
    
    if min_ != 1:
        if max_ < 1:
            if not analyzer.accepts_args():
                raise TypeError(f'`get_timeout` should accept `1` parameters, meanwhile the given callable '
                    f'expects up to `{max_!r}`, got `{get_timeout!r}`.')


def validate_close(close):
    """
    Validates the given closer.
    
    Parameters
    ----------
    close : `callable`
        The closer to validate.
    
    Raises
    ------
    TypeError
        If `close` is not `None` nor an `async-callable` accepting 1 parameter.
    """
    if close is None:
        return
    
    analyzer = CallableAnalyzer(close)
    if not analyzer.is_async():
        raise TypeError('`close` should have be `async` function.')
    
    min_, max_ = analyzer.get_non_reserved_positional_parameter_range()
    if min_ > 2:
        raise TypeError(f'`close` should accept `2` parameters, meanwhile the given callable expects at '
            f'least `{min_!r}`, got `{close!r}`.')
    
    if min_ != 2:
        if max_ < 2:
            if not analyzer.accepts_args():
                raise TypeError(f'`close` should accept `2` parameters, meanwhile the given callable expects '
                    f'up to `{max_!r}`, got `{close!r}`.')


def validate_init(init):
    """
    Validates the given initializer.
    
    Parameters
    ----------
    init : `callable`
        The initializer to validate.
    
    Raises
    ------
    TypeError
        If `init` is not `None` or is `async-callable` or accepts less than `3` parameters.
    """
    if (init is None) or (init is object.__init__):
        return
    
    analyzer = CallableAnalyzer(init)
    if analyzer.is_async():
        raise TypeError('`close` should have be `async` function.')
    
    min_, max_ = analyzer.get_non_reserved_positional_parameter_range()
    if min_ < 3:
        raise TypeError(f'`close` should accept `3` parameters, meanwhile the given callable expects at '
            f'least `{min_!r}`, got `{init!r}`.')


class MenuStructure:
    """
    A special object to store special methods of a menu.
    
    Attributes
    ----------
    check : `None` or `function`
        The function to call when checking whether an event should be called.
        
        Should accept the following parameters:
        
        +-----------+-----------------------+
        | Name      | Type                  |
        +===========+=======================+
        | self      | ``Menu``              |
        +-----------+-----------------------+
        | event     | ``InteractionEvent``  |
        +-----------+-----------------------+
        
        > ``ReactionDeleteEvent`` is only given, when the client has no `manage_messages` permission.
        
        Should return the following values:
        +-------------------+-----------+
        | Name              | Type      |
        +===================+===========+
        | should_process    | `bool`    |
        +-------------------+-----------+
    
    get_timeout : `None` or `Function`
        Return the time after the menu should be closed.
        
        > Define it as non-positive to never timeout. Not recommended.
        
        Should accept the following parameters:
        
        +-----------+-----------+
        | Name      | Type      |
        +===========+===========+
        | self      | ``Menu``  |
        +-----------+-----------+
        
        
        Should return the following parameter:
        
        +---------------+-----------+
        | Name          | Type      |
        +===============+===========+
        | timeout       | `int`     |
        +---------------+-----------+
    
    close : `None` or `CoroutineFunction`
        Function to call when the menu is closed.
        
        Should accept the following parameters:
        
        +-----------+---------------------------+
        | Name      | Type                      |
        +===========+===========================+
        | self      | ``Menu``                  |
        +-----------+---------------------------+
        | exception | `None` or `BaseException` |
        +-----------+---------------------------+
    
    init : `None` or `Function`
        Initializer function.
    
        Should accept the following parameters:
        
        +-------------------+---------------------------+
        | Name              | Type                      |
        +===================+===========================+
        | self              | ``Menu``                  |
        +-------------------+---------------------------+
        | interaction_event | ``InteractionEvent``      |
        +-------------------+---------------------------+
        | *args             | Positional parameters     |
        +-------------------+---------------------------+
        | **kwargs          | Keyword parameters        |
        +-------------------+---------------------------+
    
    initial_invoke : `None` or `CoroutineFunction`
        Function to generate the default page of the menu.
        
        Should accept the following parameters:
        
        +-----------+-----------+
        | Name      | Type      |
        +===========+===========+
        | self      | ``Menu``  |
        +-----------+-----------+
        
    is_final : `bool`
        Whether the Menu has all the required fields fulfilled.
    
    invoke : `None` or `CoroutineFunction`
        The function call for result when invoking the menu.
        
        Should accept the following parameters:
        
        +-----------+-----------------------+
        | Name      | Type                  |
        +===========+=======================+
        | self      | ``Menu``              |
        +-----------+-----------------------+
        | event     | ``InteractionEvent``  |
        +-----------+-----------------------+
        
        Should return the following parameter:
        
        +---------------+-----------+
        | Name          | Type      |
        +===============+===========+
        | should_edit   | `bool`    |
        +---------------+-----------+
    """
    __slots__ = ('check', 'close', 'get_timeout', 'init', 'is_final', 'initial_invoke', 'invoke')
    
    def __new__(cls, class_attributes):
        """
        Creates a new menu structure instance from the given class attributes dictionary.
        
        Parameters
        ----------
        class_attributes : `dict` of (`str`, `Any`) items
            Class attributes of a type.
        
        Raises
        ------
        TypeError
            - If `check` is not `None` neither a non-async function.
            - If `check` accepts less or more than `2` parameters.
            - If `invoke` is not `None` nor an `async-callable`.
            - If `invoke` accepts less or more than `2` parameters.
            - If `initial_invoke` is not `None`, nor an `async-callable`.
            - If `initial_invoke` accepts less or more than `1` parameter.
            - If `get_timeout` is not a non-async callable.
            - If `get_timeout` accepts less or more than `1` parameter.
            - If `close` is neither `None` nor `async-callable`.
            - If `close` accepts more or less than `2` parameters.
            - If `init` is not `None` or is `async-callable`.
            - If `init` accepts less than `3` parameters.
        """
        self = class_attributes.get('_menu_structure', None)
        if (self is not None) and (type(self) is cls):
            return self
        
        check = class_attributes.get('check', None)
        validate_check(check)
        
        invoke = class_attributes.get('invoke', None)
        validate_invoke(invoke)
        
        initial_invoke = class_attributes.get('initial_invoke', None)
        validate_initial_invoke(initial_invoke)
        
        get_timeout = class_attributes.get('get_timeout', None)
        validate_get_timeout(get_timeout)
        
        close = class_attributes.get('close', None)
        validate_close(close)
        
        init = class_attributes.get('__init__', None)
        validate_init(init)
        
        if (invoke is None) or (initial_invoke is None):
            is_final = False
        else:
            is_final = True
        
        self = object.__new__(cls)
        self.is_final = is_final
        self.check = check
        self.close = close
        self.init = init
        self.initial_invoke = initial_invoke
        self.invoke = invoke
        self.get_timeout = get_timeout
        
        return self
    
    
    def __repr__(self):
        """Returns the user menu structure's representation."""
        repr_parts = [
            '<',
            self.__class__.__name__,
        ]
        
        check = self.check
        if (check is None):
            field_added = False
        else:
            field_added = True
            
            repr_parts.append(' check=')
            repr_parts.append(repr(check))
        
        close = self.close
        if (close is not None):
            if field_added:
                repr_parts.append(', ')
            else:
                field_added = True
            
            repr_parts.append(' close=')
            repr_parts.append(repr(close))
        
        init = self.init
        if (init is not None):
            if field_added:
                repr_parts.append(', ')
            else:
                field_added = True
            
            repr_parts.append(' init=')
            repr_parts.append(repr(init))
        
        initial_invoke = self.initial_invoke
        if (initial_invoke is not None):
            if field_added:
                repr_parts.append(', ')
            else:
                field_added = True
            
            repr_parts.append(' initial_invoke=')
            repr_parts.append(repr(initial_invoke))
        
        invoke = self.invoke
        if (invoke is not None):
            if field_added:
                repr_parts.append(', ')
            else:
                field_added = True
            
            repr_parts.append(' invoke=')
            repr_parts.append(repr(invoke))
        
        get_timeout = self.get_timeout
        if (get_timeout is not None):
            if field_added:
                repr_parts.append(', ')
            
            repr_parts.append(' get_timeout=')
            repr_parts.append(repr(get_timeout))
        
        repr_parts.append('>')
        
        return ''.join(repr_parts)
    
    
    def merge(self, other):
        """
        Merges the two menu structure.
        
        Parameters
        ----------
        other : ``MenuStructure``
            The other menu structure to merge into self creating a new one.
        
        Returns
        -------
        new : ``MenuStructure``
        """
        check = self.check
        if check is None:
            check = other.check
        
        close = self.close
        if close is None:
            close = other.close
        
        init = self.init
        if init is None:
            init = other.init
        
        initial_invoke = self.initial_invoke
        if initial_invoke is None:
            initial_invoke = other.initial_invoke
        
        invoke = self.invoke
        if invoke is None:
            invoke = other.invoke
        
        get_timeout = self.get_timeout
        if get_timeout is None:
            get_timeout = other.get_timeout
        
        if (invoke is None) or (initial_invoke is None):
            is_final = False
        else:
            is_final = True
        
        new = object.__new__(type(self))
        new.check = check
        new.close = close
        new.init = init
        new.initial_invoke = initial_invoke
        new.invoke = invoke
        new.get_timeout = get_timeout
        new.is_final = is_final
        
        return new


def _iter_attributes(class_parents, class_attributes):
    """
    Iterates over the given class's attributes and the given attributes.
    
    Parameters
    ----------
    class_parents : `tuple` of `type`
        Parent classes.
    class_attributes : `dict` of (`str`, `Any`) items
        Class attributes of the source type.
    
    Yields
    ------
    attribute_name : `str`
        An attribute's name.
    attribute_value : `Any`
        An attribute's value.
    """
    for class_parent in reversed(class_parents):
        yield from class_parent.__dict__.items()
    
    yield from class_attributes.items()


DISALLOWED_MENU_ATTRIBUTE_NAMES = (
    '__new__',
    '__call__',
    'cancel',
    '_canceller_function',
    '_handle_close_exception',
    'content',
    'embed',
    'allowed_mentions',
    'components',
)

Menu = None

class MenuType(type):
    """
    Meta type for ``Menu`` instances.
    """
    def __new__(cls, class_name, class_parents, class_attributes):
        """
        Creates a Discord entity type. Subclass ``DiscordEntity`` instead of using this class directly as a metaclass.
        
        Parameters
        ----------
        class_name : `str`
            The created class's name.
        class_parents : `tuple` of `type` instances
            The superclasses of the creates type.
        class_attributes : `dict` of (`str`, `Any`) items
            The class attributes of the created type.
        
        Returns
        -------
        type : ``MenuType`` instance
        """
        if (Menu is not None):
            for attribute_name in DISALLOWED_MENU_ATTRIBUTE_NAMES:
                menu_attribute = getattr(Menu, attribute_name)
                try:
                    actual_attribute = class_attributes[attribute_name]
                except KeyError:
                    class_attributes[attribute_name] = menu_attribute
                else:
                    if (actual_attribute is not menu_attribute):
                        raise TypeError(f'Overwriting `{attribute_name}` is disallowed.')
        
        old_menu_structure = None
        for class_parent in reversed(class_parents):
            new_menu_structure = MenuStructure(class_parent.__dict__)
            if old_menu_structure is None:
                old_menu_structure = new_menu_structure
            else:
                old_menu_structure = new_menu_structure.merge(old_menu_structure)
        
        new_menu_structure = MenuStructure(class_attributes)
        if (old_menu_structure is not None):
            new_menu_structure = new_menu_structure.merge(old_menu_structure)
        
        class_attributes['_menu_structure'] = new_menu_structure
        
        
        all_component_descriptor = {}
        
        component_descriptors = OrderedDict()
        tracked_components = OrderedDict()
        
        for attribute_name, attribute_value in _iter_attributes(class_parents, class_attributes):
            if isinstance(attribute_value, ComponentDescriptor):
                container = component_descriptors
                secondary = tracked_components
                component_hasher = ComponentSourceIdentityHasher(attribute_value._source_component)
                all_component_descriptor[component_hasher] = attribute_value
            elif isinstance(attribute_value, ComponentBase):
                container = tracked_components
                secondary = component_descriptors
            else:
                continue
            
            container[attribute_name] = attribute_value
            container.move_to_end(attribute_name)
            
            try:
                del secondary[attribute_name]
            except KeyError:
                pass
            continue
        
        component_groups = []
        component_standalone = []
        
        for attribute_name, attribute_value in tracked_components.items():
            if attribute_value.type in (ComponentType.row, ComponentType.select):
                container = component_groups
            else:
                container = component_standalone
            
            container.append((attribute_name, attribute_value))
            continue
        
        
        for component_name, component_value in component_standalone:
            component_hasher = ComponentSourceIdentityHasher(component_value)
            
            try:
                component_descriptor = all_component_descriptor[component_hasher]
            except KeyError:
                component_descriptor = ComponentDescriptor(component_value, None)
                all_component_descriptor[component_hasher] = component_descriptor
            
            component_descriptors[component_name] = component_descriptor
        
        for component_name, component_value in component_groups:
            component_hasher = ComponentSourceIdentityHasher(component_value)
            
            try:
                component_descriptor = all_component_descriptor[component_hasher]
            except KeyError:
                sub_component_relations = None
                sub_component_descriptors = None
                for sub_component in component_value._iter_direct_sub_components():
                    sub_component_hasher = ComponentSourceIdentityHasher(sub_component)
                    try:
                        sub_component_descriptor = all_component_descriptor[sub_component_hasher]
                    except KeyError:
                        sub_component_descriptor = ComponentDescriptor(sub_component, None)
                        all_component_descriptor[sub_component_hasher] = sub_component_descriptor
                    
                    if (sub_component_descriptors is None):
                        sub_component_descriptors = []
                    sub_component_descriptors.append(sub_component_descriptor)
                    
                    if (sub_component_relations is None):
                        sub_component_relations = {}
                    sub_component_relations[sub_component] = sub_component_descriptor
                
                if (sub_component_relations is not None):
                    component_value._replace_direct_sub_components(sub_component_relations)
                
                component_descriptor = ComponentDescriptor(component_value, sub_component_descriptors)
                all_component_descriptor[component_hasher] = component_descriptor
            
            component_descriptors[component_name] = component_descriptor
        
        for attribute_name, attribute_value in component_descriptors.items():
            class_attributes[attribute_name] = attribute_value
        
        if '__slots__' not in class_attributes:
            class_attributes['__slots__'] = ('__dict__', )
        
        return type.__new__(cls, class_name, class_parents, class_attributes)


class Menu(metaclass=MenuType):
    """
    Base class for custom component based menus.
    
    Attributes
    ----------
    _allowed_mentions : ``Ellipsis``, `None` or `list` of `Any`
        The used allowed mentions when editing the respective message.
    _canceller : None` or `CoroutineFunction`
        Canceller set as `._canceller_function``, meanwhile the gui is not cancelled.
    _components : `None` or `tuple` of ``ComponentBase``
        Rendered components of the menu.
    _component_proxy_cache : `dict` of (`int`, ``ComponentProxy``) items
        A dictionary of component proxy identifiers and component proxies.
    _gui_state : `int`
        The gui's state.
        
        Can be any of the following:
        
        +-------------------------------+-------+
        | Respective name               | Value |
        +===============================+=======+
        | GUI_STATE_NONE                | 0     |
        +===============================+=======+
        | GUI_STATE_READY               | 1     |
        +-------------------------------+-------+
        | GUI_STATE_EDITING             | 2     |
        +-------------------------------+-------+
        | GUI_STATE_CANCELLING          | 3     |
        +-------------------------------+-------+
        | GUI_STATE_CANCELLED           | 4     |
        +-------------------------------+-------+
        | GUI_STATE_SWITCHING_CONTEXT   | 5     |
        +-------------------------------+-------+
    _timeouter : `None` or ``Timeouter``
        Executes the timeout feature on the menu.
    _tracked_changes : `dict` of (`str`, `Any`) items
        The tracked changes by parameter name.
    channel : ``ChannelTextBase`` instance
        The channel where the menu is executed.
    client : ``Client``
        The executor client instance.
    message : `None` or ``Message``
        The message which executes the menu.
    
    Class Attributes
    ----------------
    _menu_structure : ``MenuStructure``
        Factorized methods added by the user used by the menu itself.
    """
    __slots__ = ('_allowed_mentions', '_canceller', '_component_proxy_cache', '_gui_state', '_timeouter',
        '_tracked_changes', 'channel', 'client', 'message', '_components',)
    
    async def __new__(cls, client, target, *args, **kwargs):
        """
        Creates a new menu instance.
        
        Parameters
        ----------
        client : ``Client``
            The client instance whi will execute the action.
        target : ``InteractionEvent``, ``Message``, ``ChannelTextBase``, `int`
            The event to respond to, or the channel to send the message at, or the message to edit.
        *args : Positional parameters
            Additional positional parameters.
        **kwargs : Keyword parameters
            Additional keyword parameters.
        
        Raises
        ------
        TypeError
            If `interaction_event` was not given neither as ``InteractionEvent``, ``Message``, ``ChannelTextBase``,
             nor as instance.
        RuntimeError
            If `interaction_event` is given as ``InteractionEvent``, but it's client cannot be detected.
        """
        menu_structure = cls._menu_structure
        if not menu_structure.is_final:
            raise RuntimeError(f'{cls.__class__.__name__} has not every required fields fulfilled. Required fields: '
                f'initial_invoke, invoke')
        
        # use goto
        while True:
            if isinstance(target, InteractionEvent):
                target_message = target.message # Should be `None`
                interaction_event = target
                is_interaction = True
                break
                
            if isinstance(target, Message):
                target_channel_id = target.channel_id
                target_message = target
                is_interaction = False
                break
            
            if isinstance(target, ChannelTextBase):
                target_channel_id = target.id
                target_message = None
                is_interaction = False
                break
            
            if isinstance(target, int):
                target_channel_id = target
                target_message = None
                is_interaction = False
                break
                
                
            raise TypeError(f'`target` can be given as `{InteractionEvent.__name__}`, '
                f'`{Message.__name__}`, `{ChannelTextBase.__name__}` or as `int` instance, got '
                f'{target.__class__.__name__}.')
        
        
        self = object.__new__(cls)
        self._canceller = None
        self.message = target_message
        self.client = client
        self._gui_state = GUI_STATE_NONE
        self._timeouter = None
        self._tracked_changes = {}
        self._allowed_mentions = None
        self._component_proxy_cache = {}
        self._components = None
        
        init = menu_structure.init
        if (init is not None):
            init(self, client, target, *args, **kwargs)
        
        await menu_structure.initial_invoke(self)
        
        tracked_changes = self._tracked_changes
        if not tracked_changes:
            raise RuntimeError(f'{cls.__class__.__name__}\'s `initial_invoke` method: '
                f'{menu_structure.initial_invoke!r} did not change any parameters.')
        
        kwargs = tracked_changes.copy()
        tracked_changes.clear()
        
        allowed_mentions = self._allowed_mentions
        if (allowed_mentions is not None):
            kwargs['allowed_mentions'] = allowed_mentions
        
        kwargs['components'] = self._components
        
        if is_interaction:
            interaction_event_type = interaction_event.type
            if interaction_event_type is INTERACTION_TYPE_APPLICATION_COMMAND:
                if target_message is None:
                    if interaction_event.is_unanswered():
                        await client.interaction_application_command_acknowledge(interaction_event)
                    
                    target_message = await client.interaction_followup_message_create(interaction_event, **kwargs)
                else:
                    await client.interaction_response_message_edit(interaction_event, **kwargs)
            
            elif interaction_event_type is INTERACTION_TYPE_MESSAGE_COMPONENT:
                await client.interaction_component_message_edit(interaction_event, **kwargs)
            
            else:
                # nani desu ka?
                return
        
        else:
            if target_message is None:
                target_message = await client.message_create(target_channel_id, **kwargs)
            else:
                await client.message_edit(target_message, **kwargs)
        
        self.message = target_message
        
        get_timeout = menu_structure.get_timeout
        if (get_timeout is None):
            timeout = -1.0
        else:
            timeout = get_timeout(self)
        
        if (timeout > 0.0):
            self._timeouter = Timeouter(self, timeout)
        
        self._gui_state = GUI_STATE_READY
        self._canceller = cls._canceller_function
        
        client.slasher.add_component_interaction_waiter(target_message, self)
        
        return self
    
    
    async def __call__(self, interaction_event):
        """
        Processes a received component interaction event.
        
        This method is a coroutine.
        
        Parameters
        ----------
        interaction_event : ``InteractionEvent``
            Component interaction event.
        """
        client = self.client
        gui_state = self._gui_state
        if (gui_state == GUI_STATE_EDITING) or (GUI_STATE_EDITING == GUI_STATE_CANCELLING):
            await client.interaction_component_acknowledge(interaction_event)
        
        if (gui_state != GUI_STATE_READY):
            return
        
        check = self._menu_structure.check
        if (check is not None):
            try:
                should_process = check(self, interaction_event)
            except BaseException as err:
                self.cancel(err)
                return
            
            if not should_process:
                await client.interaction_component_acknowledge(interaction_event)
                return
        
        self._gui_state = GUI_STATE_EDITING
        
        invoke = self._menu_structure.invoke
        try:
            should_edit = await invoke(self, interaction_event)
        except BaseException as err:
            self.cancel(err)
            return
        
        if should_edit:
            tracked_changes = self._tracked_changes
            kwargs = tracked_changes.copy()
            tracked_changes.clear()
            
            allowed_mentions = self._allowed_mentions
            if (allowed_mentions is not None):
                kwargs['allowed_mentions'] = allowed_mentions
            
            try:
                await client.interaction_component_message_edit(interaction_event, **kwargs, components=self._components)
            except BaseException as err:
                self.cancel(err)
                return
            
            
            get_timeout = self._menu_structure.get_timeout
            if (get_timeout is None):
                timeout = -1.0
            else:
                timeout = get_timeout(self)
            
            timeouter = self._timeouter
            if (timeout > 0.0):
                if timeouter is None:
                    self._timeouter = Timeouter(self, timeout)
                else:
                    timeouter.set_timeout(timeout)
            else:
                if (timeouter is not None):
                    self._timeouter = None
                    timeouter.cancel()
        
        else:
            try:
                await client.interaction_component_acknowledge(interaction_event)
            except BaseException as err:
                self.cancel(err)
                return
        
        if self._gui_state == GUI_STATE_EDITING:
            self._gui_state = GUI_STATE_READY
    
    
    def cancel(self, exception=None):
        """
        Cancels the menu with the given exception.
        
        Parameters
        ----------
        exception : `None` or ``BaseException`` instance, Optional
            Exception to cancel the pagination with. Defaults to `None`
        
        Returns
        -------
        canceller_task : `None` or ``Task``
        """
        if self._gui_state in (GUI_STATE_READY, GUI_STATE_EDITING, GUI_STATE_CANCELLING):
            self._gui_state = GUI_STATE_CANCELLED
        
        canceller = self._canceller
        if canceller is None:
            return
        
        self._canceller = None
        
        timeouter = self._timeouter
        if (timeouter is not None):
            timeouter.cancel()
        
        return Task(canceller(self, exception), KOKORO)
    
    
    async def _canceller_function(self, exception):
        """
        Cancels the gui state, saving the current game if needed.
        
        This method is a coroutine.
        
        Parameters
        ----------
        exception : `None`, ``BaseException``
        """
        client = self.client
        message = self.message
        
        client.slasher.remove_component_interaction_waiter(message, self)
        
        
        if self._gui_state == GUI_STATE_SWITCHING_CONTEXT:
            # the message is not our, we should not do anything with it.
            return
        
        self._gui_state = GUI_STATE_CANCELLED
        
        close = self._menu_structure.close
        
        try:
            if (close is None):
                handled = await self._handle_close_exception(exception)
            else:
                await close(self, exception)
                handled = True
        except BaseException as err:
            await client.events.error(client, f'{self!r}._canceller_function', err)
        else:
            if not handled:
                await client.events.error(client, f'{self!r}.cancel', exception)
    
    
    async def _handle_close_exception(self, exception):
        """
        Handles close exception if any.
        
        This method is a coroutine.
        
        Parameters
        ----------
        exception : `None` or `BaseException`
            The close exception to handle.
        
        Returns
        -------
        handled : `bool`
            Whether the exception was handled.
        """
        if exception is None:
            return True
        
        client = self.client
        message = self.message
        
        if isinstance(exception, CancelledError):
            try:
                await client.message_delete(message)
            except BaseException as err:
                if isinstance(err, ConnectionError):
                    # no internet
                    return True
                
                if isinstance(err, DiscordException):
                    if err.code in (
                            ERROR_CODES.unknown_channel, # channel deleted
                            ERROR_CODES.unknown_message, # message deleted
                            ERROR_CODES.missing_access, # client removed
                                ):
                        return True
                
                await client.events.error(client, f'{self!r}._handle_close_exception', err)
            
            return True
        
        if isinstance(exception, TimeoutError):
            if (message.components is not None):
                try:
                    await client.message_edit(message, components=None)
                except BaseException as err:
                    if isinstance(err, ConnectionError):
                        # no internet
                        return True
                    
                    if isinstance(err, DiscordException):
                        if err.code in (
                                ERROR_CODES.unknown_message, # message deleted
                                ERROR_CODES.unknown_channel, # channel deleted
                                ERROR_CODES.missing_access, # client removed
                                ERROR_CODES.missing_permissions, # permissions changed meanwhile
                                    ):
                            return True
                    
                    await client.events.error(client, f'{self!r}._handle_close_exception', err)
            
            return True
        
        if isinstance(exception, PermissionError):
            return True
        
        return False
    
    
    @property
    def content(self):
        """
        A get-set-del property to modify the menu's content.
        """
        message = self.message
        if message is None:
            content = None
        else:
            content = message.content
        
        return content
    
    
    @content.setter
    def content(self, content):
        if content is None:
            change_name = 'content'
            change_value = ''
        elif isinstance(content, str):
            change_name = 'content'
            change_value = content
        elif isinstance(content, EmbedBase):
            change_name = 'embed'
            change_value = content
        else:
            change_name = 'content'
            change_value = str(content)
            if len(change_value) > 2000:
                change_value = change_value[:2000]
        
        self._tracked_changes[change_name] = change_value
    
    @content.deleter
    def content(self):
        self._tracked_changes['content'] = None
    
    
    @property
    def embed(self):
        """
        A get-set-del property to modify the menu's embed.
        """
        message = self.message
        if message is None:
            embed = None
        else:
            embeds = message.embeds
            if embeds is None:
                embed = None
            else:
                embed = embeds[0]
        
        return embed
    
    @embed.setter
    def embed(self, embed):
        if (embed is not None) and (not isinstance(embed, EmbedBase)):
            raise TypeError(f'`embed` can be given as `None` or as `{EmbedBase.__name__}` instance, got '
                f'{embed.__class__.__name__}.')
        
        self._tracked_changes['embed'] =  embed
    
    @embed.deleter
    def embed(self):
        self._tracked_changes['embed'] = None
    
    
    @property
    def allowed_mentions(self):
        """
        A get-set-del property to modify the menu's message's allowed mentions.
        """
        allowed_mentions = self._allowed_mentions
        if allowed_mentions is None:
            allowed_mentions_output = AllowedMentionProxy()
        else:
            allowed_mentions_output = allowed_mentions.copy()
        
        return allowed_mentions_output
    
    @allowed_mentions.setter
    def allowed_mentions(self, allowed_mentions_input):
        if allowed_mentions_input is None:
            allowed_mentions = AllowedMentionProxy()
        elif isinstance(allowed_mentions_input, AllowedMentionProxy):
            allowed_mentions = allowed_mentions_input.copy()
        else:
            allowed_mentions = AllowedMentionProxy(allowed_mentions_input)
        
        self._allowed_mentions = allowed_mentions
    
    @allowed_mentions.deleter
    def allowed_mentions(self):
        self._allowed_mentions = None
    
    
    @property
    def components(self):
        """
        A get-set-del property to modify the menu's components.
        """
        return tuple(self._components)
    
    
    @components.setter
    def components(self, raw_components):
        components = []
        all_component_descriptor = None
        
        if raw_components is None:
            pass
        
        elif isinstance(raw_components, ComponentProxy):
            if raw_components.type is ComponentType.row:
                components.append(raw_components)
            else:
                
                component_descriptor = ComponentDescriptorState(
                    ComponentRow(raw_components._source_component),
                    [raw_components],
                )
                component_proxy = component_descriptor.get_component_proxy(self)
                
                components.append(component_proxy)
        
        elif isinstance(raw_components, ComponentBase):
            raw_component = raw_components
            sub_component_descriptors = None
            sub_component_relations = None
            for sub_component in raw_component._iter_direct_sub_components():
                sub_component_hasher = ComponentSourceIdentityHasher(sub_component)
                
                if all_component_descriptor is None:
                    all_component_descriptor = {}
                
                try:
                    sub_component_descriptor = all_component_descriptor[sub_component_hasher]
                except KeyError:
                    sub_component_descriptor = ComponentDescriptor(sub_component, None)
                    all_component_descriptor[sub_component_hasher] = sub_component_descriptor
                
                if sub_component_descriptors is None:
                    sub_component_descriptors = []
                sub_component_descriptors.append(sub_component_descriptor)
                
                if sub_component_relations is None:
                    sub_component_relations = {}
                sub_component_relations[sub_component] = sub_component_descriptor
            
            if (sub_component_relations is not None):
                raw_component._replace_direct_sub_components(sub_component_relations)
            
            component_descriptor = ComponentDescriptorState(
                ComponentRow(raw_component),
                sub_component_descriptors,
            )
            
            component_proxy = component_descriptor.get_component_proxy(self)
            
            components.append(component_proxy)
        
        elif isinstance(raw_components, (tuple, list)):
            
            for raw_sub_component in raw_components:
                if isinstance(raw_sub_component, ComponentProxy):
                    if raw_sub_component.type is ComponentType.row:
                        components.append(raw_sub_component)
                    else:
                        
                        component_descriptor = ComponentDescriptorState(
                            ComponentRow(raw_sub_component._source_component),
                            [raw_sub_component],
                        )
                        component_proxy = component_descriptor.get_component_proxy(self)
                        
                        components.append(component_proxy)
                    
                elif isinstance(raw_sub_component, ComponentBase):
                    component_hasher = ComponentSourceIdentityHasher(raw_sub_component)
                    
                    if all_component_descriptor is None:
                        all_component_descriptor = {}
                    
                    try:
                        component_descriptor = all_component_descriptor[component_hasher]
                    except KeyError:
                        sub_component_relations = None
                        sub_component_descriptors = None
                        for sub_component in raw_sub_component._iter_direct_sub_components():
                            sub_component_hasher = ComponentSourceIdentityHasher(sub_component)
                            
                            if all_component_descriptor is None:
                                all_component_descriptor = {}
                            
                            try:
                                sub_component_descriptor = all_component_descriptor[sub_component_hasher]
                            except KeyError:
                                sub_component_descriptor = ComponentDescriptor(sub_component, None)
                                all_component_descriptor[sub_component_hasher] = sub_component_descriptor
                            
                            if sub_component_descriptors is None:
                                sub_component_descriptors = []
                                sub_component_descriptors.append(sub_component_descriptor)
                            if sub_component_relations is None:
                                sub_component_relations = {}
                            sub_component_relations[sub_component] = sub_component_descriptor
                        
                        if (sub_component_relations is not None):
                            raw_sub_component._replace_direct_sub_components(sub_component_relations)
                        
                        
                        component_descriptor = ComponentDescriptorState(
                            raw_sub_component,
                            sub_component_descriptors,
                        )
                        
                        all_component_descriptor[component_hasher] = component_descriptor
                    

                    component_proxy = component_descriptor.get_component_proxy(self)
                    components.append(component_proxy)
                
                elif isinstance(raw_sub_component, (tuple, list)):
                    sub_component_descriptors = []
                    for raw_nested_sub_component in raw_sub_component:
                        if isinstance(raw_nested_sub_component, ComponentProxy):
                            if raw_nested_sub_component.type is ComponentType.row:
                                raise TypeError(f'Cannot double-nest row components.')
                            
                            sub_component_descriptors.append(raw_nested_sub_component)
                        
                        elif isinstance(raw_nested_sub_component, ComponentBase):
                            if raw_nested_sub_component.type is ComponentType.row:
                                raise TypeError(f'Cannot double-nest row components.')
                            
                            sub_component_hasher = ComponentSourceIdentityHasher(raw_nested_sub_component)
                            
                            if all_component_descriptor is None:
                                all_component_descriptor = {}
                            
                            try:
                                sub_component_descriptor = all_component_descriptor[sub_component_hasher]
                            except KeyError:
                                sub_component_descriptor = ComponentDescriptor(raw_nested_sub_component, None)
                                all_component_descriptor[sub_component_hasher] = sub_component_descriptor
                            
                            sub_component_descriptors.append(sub_component_descriptor)
                        
                        else:
                            raise TypeError(f'Nested-sub components can be either `{ComponentProxy.__name__}` or  '
                                f'`{ComponentBase.__name__}`, got {raw_nested_sub_component.__class__.__name__}.')
                    
                    component_descriptor = ComponentDescriptorState(
                        ComponentRow(*raw_components),
                        sub_component_descriptors,
                    )
                    
                    component_proxy = component_descriptor.get_component_proxy(self)
                    
                    components.append(component_proxy)
        
        else:
            raise TypeError(f'`raw_components` can be `None`, `{ComponentProxy.__name__}`, `{ComponentBase.__name__}`, '
                f'(`list`, `tuple`) of repeat, no triple nesting, got {raw_components.__class__.__name__}.')
        
        if not components:
            components = None
        
        self._components = components
    
    @components.deleter
    def components(self):
        self._components = None



__all__ = ('Emoji',)

from ...backend.export import include, export

from ..bases import DiscordEntity, id_sort_key
from ..core import EMOJIS, GUILDS, BUILTIN_EMOJIS, UNICODE_TO_EMOJI
from ..utils import id_to_datetime, DISCORD_EPOCH_START, DATETIME_FORMAT_CODE
from ..user import User, ZEROUSER
from ..preconverters import preconvert_str, preconvert_bool, preconvert_snowflake
from ..role import create_partial_role_from_id, Role
from ..bases import instance_or_id_to_instance, iterable_of_instance_or_id_to_snowflakes, instance_or_id_to_snowflake
from ..http import urls as module_urls

Client = include('Client')
Guild = include('Guild')

UNICODE_EMOJI_LIMIT = 1<<21


@export
class Emoji(DiscordEntity, immortal=True):
    """
    Represents a Discord emoji. It can be custom or builtin (unicode) emoji as well. Builtin emojis are loaded when the
    module is imported and they are stores at `BUILTIN_EMOJIS` dictionary. At `BUILTIN_EMOJIS` the keys are the
    emoji's names, so it is easy to access any Discord unicode emoji like that.
    
    Custom emojis are loaded with ``Guild``-s on startup, but new partial custom emojis can be created later as well,
    when a ``Message`` receives any reaction.
    
    Attributes
    ----------
    id : `int`
        Unique identifier of the emoji.
    animated : `bool`
        Whether the emoji is animated.
    available : `bool`
        Whether the emoji is available.
    guild_id : `int`
        The emoji's guild's identifier.
    managed : `bool`
        Whether the emoji is managed by an integration.
    name : `int`
        The emoji's name.
    role_ids : `None` or `tuple` of `int`
        A set of role identifiers for which the custom emoji is whitelisted to. If the emoji is not limited for specific roles,
        then this value is set to `None`. If the emoji is a builtin (unicode) emoji, then this attribute is set to
        `None` as  well.
    unicode : `None` or `str`
        At the case of custom emojis this attribute is always `None`, but at the case of builtin (unicode) emojis this
        attribute stores the emoji's unicode representation.
    user : ``ClientUserBase``
        The creator of the custom emoji. The emoji must be requested from Discord's API, or it's user will be just
        the default `ZEROUSER`.
    
    Class Attributes
    ----------------
    _last_unicode_id : `int`
        The most recently created unicode emoji's identifier.
    
    See Also
    --------
    - ``create_partial_emoji`` : A function to create an emoji object from partial emoji data.
    - ``parse_emoji`` : Parses a partial emoji object out from text.
    """
    __slots__ = ('animated', 'available', 'guild_id', 'managed', 'name', 'require_colons', 'role_ids', 'unicode',
        'user', )
    
    _last_unicode_id = 0
    
    def __new__(cls, data, guild):
        """
        Creates a new emoji object from emoji data included with it's guild's. If the emoji already exists, picks that
        up instead of creating a new one.
        
        This method can not create builtin (unicode) emojis. Those are created when the library is loaded.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Emoji data received from Discord.
        guild : ``Guild``
            The guild of the emoji.
        
        Returns
        -------
        emoji : ``Emoji``
        """
        emoji_id = int(data['id'])

        try:
            self = EMOJIS[emoji_id]
        except KeyError:
            self = object.__new__(cls)
            self.id = emoji_id
            EMOJIS[emoji_id] = self
            
            try:
                user_data = data['user']
            except KeyError:
                user = ZEROUSER
            else:
                user = User(user_data)
            self.user = user
            
        else:
            # whenever we receive an emoji, it will have no user data included,
            # so it is enough if we check for user data only whenever we
            # receive emoji data from a request or so.
            if not self.partial:
                
                if self.user is ZEROUSER:
                    try:
                        user_data = data['user']
                    except KeyError:
                        pass
                    else:
                        self.user = User(user_data)
                
                return self
        
        name = data['name']
        if name is None:
            name = ''
        
        self.name = name
        self.animated = data.get('animated', False)
        self.require_colons= data.get('require_colons', True)
        self.managed = data.get('managed', False)
        self.guild_id = guild.id
        self.available = data.get('available', True)
        self.user = ZEROUSER
        
        self.unicode = None
        
        role_ids = data.get('roles', None)
        if (role_ids is None) or (not role_ids):
            role_ids = None
        else:
            role_ids = tuple(sorted(int(role_id)for role_id in role_ids))
        
        self.role_ids = role_ids
        
        return self
    
    
    @classmethod
    def precreate(cls, emoji_id, **kwargs):
        """
        Precreates the emoji by creating a partial one with the given parameters. When the emoji is loaded
        the precrated one will be picked up. If an already existing emoji would be precreated, returns that
        instead and updates that only, if that is partial.
        
        Parameters
        ----------
        emoji_id : `snowflake`
            The emoji's id.
        **kwargs : keyword parameters
            Additional predefined attributes for the emoji.
        
        Other Parameters
        ----------------
        animated : `bool`, Optional (Keyword only)
             The emoji's ``.animated``.
        available : `bool`, Optional (Keyword only)
             The emoji's ``.available``.
        guild_id : ``Guild``, `int`, Optional (Keyword only)
             The emoji's ``.guild_id``.
        name : `str`, Optional (Keyword only)
            The emoji's ``.name``. Can be between length `2` and `32`.
        require_colons : `bool`, Optional (Keyword only)
             The emoji's ``.require_colons``.
        roles : `None` or `iterable` of (``Role``, `int`), Optional (Keyword only)
            The emoji's ``.roles``.
        user : ``ClientUserBase`, `int`, Optional (Keyword only)
            The emoji's ``.user``.
        
        Returns
        -------
        self : ``Emoji``
        
        Raises
        ------
        TypeError
            If any parameter's type is bad or if unexpected parameter is passed.
        ValueError
            If an parameter's type is good, but it's value is unacceptable.
        """
        emoji_id = preconvert_snowflake(emoji_id, 'emoji_id')
        
        if kwargs:
            processable = []
            
            try:
                name = kwargs.pop('name')
            except KeyError:
                pass
            else:
                name = preconvert_str(name, 'name', 2, 32)
                processable.append(('name', name))
            
            for attribute_name in ('animated', 'available', 'managed', 'require_colons',):
                try:
                    attribute_value = kwargs.pop(attribute_name)
                except KeyError:
                    pass
                else:
                    animated = preconvert_bool(attribute_value, attribute_name)
                    processable.append((attribute_name, animated))
            
            
            for attribute_name, attribute_type, converter in (
                ('guild_id', Guild, instance_or_id_to_snowflake),
                ('user', (User, Client), instance_or_id_to_instance),
            ):
                try:
                    attribute_value = kwargs.pop(attribute_name)
                except KeyError:
                    pass
                else:
                    attribute_value = converter(attribute_value, attribute_type, attribute_name)
                    processable.append((attribute_name, attribute_value))
            
            
            try:
                roles = kwargs.pop('roles')
            except KeyError:
                pass
            else:
                if (roles is not None):
                    role_ids = iterable_of_instance_or_id_to_snowflakes(roles, Role, 'roles')
                    if role_ids:
                        role_ids = tuple(sorted(role_ids))
                        processable.append(('role_ids', role_ids))
            
            if kwargs:
                raise TypeError(f'Unused or unsettable attributes: {kwargs}')
        
        else:
            processable = None
        
        
        try:
            self = EMOJIS[emoji_id]
        except KeyError:
            self = cls._create_empty(emoji_id)
            
            EMOJIS[emoji_id] = self
        else:
            if not self.partial:
                return self
        
        if (processable is not None):
            for name, value in processable:
                setattr(self, name, value)
        
        return self
    
    
    def __repr__(self):
        """Returns the emoji's representation."""
        return f'<{self.__class__.__name__} id={self.id}, name={self.name!r}>'
    
    
    def __format__(self, code):
        """
        Formats the emoji in a format string.
        
        Parameters
        ----------
        code : `str`
            The option on based the result will be formatted.
        
        Returns
        -------
        emoji : `str`
        
        Raises
        ------
        ValueError
            Unknown format code.
        
        Examples
        --------
        ```py
        >>> from hata import Emoji, now_as_id, BUILTIN_EMOJIS
        >>> emoji = Emoji.precreate(now_as_id(), name='nice')
        >>> emoji
        <Emoji id=712359434843586560, name='nice'>
        >>> # no code returns the emoji's emoji format as a shortcut in f string formatting.
        >>> f'{emoji}'
        '<:nice:712359434843586560>'
        >>> # 'e' stands for emoji format.
        >>> f'{emoji:e}'
        '<:nice:712359434843586560>'
        >>> # 'r' stands for reaction format.
        >>> f'{emoji:r}'
        'nice:712359434843586560'
        >>> # 'c' stands for created at.
        >>> f'{emoji:c}'
        '2020.05.19-17:42:04'
        >>> # The following works with builtin (unicode) emojis as well.
        >>> emoji = BUILTIN_EMOJIS['heart']
        >>> f'{emoji}'
        '❤️'
        >>> f'{emoji:e}'
        '❤️'
        >>> f'{emoji:r}'
        '❤️'
        >>> f'{emoji:c}'
        '2015.01.01-00:00:00'
        ```
        """
        if not code:
            return self.as_emoji
        
        if code == 'e':
            return self.as_emoji
        
        if code == 'r':
            return self.as_reaction
        
        if code == 'c':
            return self.created_at.__format__(DATETIME_FORMAT_CODE)
        
        raise ValueError(f'Unknown format code {code!r} for object of type {self.__class__.__name__!r}')
    
    
    @property
    def partial(self):
        """
        Returns whether the emoji is partial.
        
        Returns
        -------
        partial : `bool`
        """
        if (self.unicode is not None):
            return False
        
        guild_id = self.guild_id
        if not guild_id:
            return True
        
        try:
            guild = GUILDS[guild_id]
        except KeyError:
            return True
        
        return guild.partial
    
    
    def is_custom_emoji(self):
        """
        Returns whether the emoji is a custom emoji.
        
        Returns
        -------
        is_custom_emoji : `bool`
        """
        return (self.id >= UNICODE_EMOJI_LIMIT)
    
    
    def is_unicode_emoji(self):
        """
        Returns whether the emoji is a unicode emoji.
        
        Returns
        -------
        is_custom_emoji : `bool`
        """
        return (self.id < UNICODE_EMOJI_LIMIT)
    
    
    @property
    def as_reaction(self):
        """
        Returns the emoji's reaction form, which is used by the Discord API at requests when working with reactions.
        
        Returns
        -------
        as_reaction : `str`
        """
        if self.id < UNICODE_EMOJI_LIMIT:
            return self.unicode
        
        return f'{self.name}:{self.id}'
    
    
    @property
    def as_emoji(self):
        """
        Returns the emoji's emoji form. Should be used when sending an emoji within a ``Message``.
        
        Returns
        -------
        as_emoji : `str`
        """
        if self.id < UNICODE_EMOJI_LIMIT:
            return self.unicode
        
        if self.animated:
            return f'<a:{self.name}:{self.id}>'
        else:
            return f'<:{self.name}:{self.id}>'
    
    
    @property
    def created_at(self):
        """
        When the emoji was created. If the emoji is unicode emoji, then returns Discord epoch's start.
        
        Returns
        -------
        created_at : `datetime`
        """
        id_ = self.id
        if id_ > UNICODE_EMOJI_LIMIT:
            created_at = id_to_datetime(id_)
        else:
            created_at = DISCORD_EPOCH_START
        
        return created_at
    
    
    url = property(module_urls.emoji_url)
    url_as = module_urls.emoji_url_as
    
    
    def _update_attributes(self, data):
        """
        Updates the emoji with overwriting it's old attributes.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Emojis data received from Discord
        """
        self.require_colons = data.get('require_colons', True)
        self.managed = data.get('managed', False)
        
        self.animated = data.get('animated', False)
        
        name = data['name']
        if name is None:
            name = ''
        
        self.name = name
        
        role_ids = data.get('roles', None)
        if (role_ids is None) or (not role_ids):
            role_ids = None
        else:
            role_ids = tuple(sorted(int(role_id) for role_id in role_ids))
        self.role_ids = role_ids
        
        try:
            user_data = data['user']
        except KeyError:
            pass
        else:
            self.user = User(user_data)
        
        self.available = data.get('available', True)
    
    
    def _difference_update_attributes(self, data):
        """
        Updates the emoji and returns it's overwritten old attributes as a `dict` with a `attribute-name` - `old-value`
        relation.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Emoji data received from Discord.
        
        Returns
        -------
        old_attributes : `dict` of (`str`, `Any`) items
            All item in the returned dict is optional.
        
        Returned Data Structure
        -----------------------
        +-------------------+-------------------------------+
        | Keys              | Values                        |
        +===================+===============================+
        | animated          | `bool`                        |
        +-------------------+-------------------------------+
        | available         | `bool`                        |
        +-------------------+-------------------------------+
        | managed           | `bool`                        |
        +-------------------+-------------------------------+
        | name              | `int`                         |
        +-------------------+-------------------------------+
        | require_colons    | `bool`                        |
        +-------------------+-------------------------------+
        | role_ids          | `None` or `tuple` of ``Role`` |
        +-------------------+-------------------------------+
        """
        old_attributes = {}
        
        require_colons = data.get('require_colons', True)
        if self.require_colons != require_colons:
            old_attributes['require_colons'] = self.require_colons
            self.require_colons = require_colons
        
        managed = data.get('managed', False)
        if self.managed != managed:
            old_attributes['managed'] = self.managed
            self.managed = managed
        
        animated = data.get('animated', False)
        if self.animated != animated:
            old_attributes['animated'] = self.animated
            self.animated = animated
        
        name = data['name']
        if name is None:
            name = ''
        if self.name != name:
            old_attributes['name'] = self.name
            self.name = name
        
        role_ids = data.get('roles', None)
        if (role_ids is None) or (not role_ids):
            role_ids = None
        else:
            role_ids = tuple(sorted(int(role_id) for role_id in role_ids))
        
        if self.role_ids != role_ids:
            old_attributes['role_ids'] = self.role_ids
            self.role_ids = role_ids
        
        try:
            user_data = data['user']
        except KeyError:
            pass
        else:
            self.user = User(user_data)
        
        available = data.get('available', True)
        if self.available != available:
            old_attributes['available'] = self.available
            self.available = available
        
        return old_attributes
    
    
    @classmethod
    def _create_partial(cls, emoji_id, name, animated):
        """
        Creates a new emoji from the given partial data.
        
        Parameters
        ----------
        emoji_id : `int`
            The emoji's identifier.
        name : `str`
            The emoji's name.
        animated : `bool`
            Whether the emoji is animated.
        
        Returns
        -------
        emoji : ``Emoji``
        """
        emoji_id = int(emoji_id)
        
        try:
            self = EMOJIS[emoji_id]
            if self.partial:
                self.name = name
                self.animated = animated
        except KeyError:
            self = cls._create_empty(emoji_id)
            self.name = name
            self.animated = animated
            EMOJIS[emoji_id] = self
        
        return self
    
    
    @classmethod
    def _create_empty(cls, emoji_id):
        """
        Creates an empty emoji with the given identifier.
        
        Parameters
        ----------
        emoji_id : `int`
            The emoji's identifier.
        
        Returns
        -------
        self : ``Emoji``
            The created emoji.
        """
        self = object.__new__(cls)
        self.id = emoji_id
        self.animated = False
        self.name = ''
        self.unicode = None
        self.guild_id = 0
        self.available = True
        self.require_colons = True
        self.managed = False
        self.user = ZEROUSER
        self.role_ids = None
        return self
    
    
    @classmethod
    def _create_unicode(cls, name, unicode, aliases):
        """
        Creates a new unicode emoji with the given identifier.
        
        Parameters
        ----------
        name : `str`
            The emoji's name.
        unicode : `str`
            The emoji's unicode value.
        aliases : `tuple` of `str`
            Emoji name aliases.
        
        Returns
        -------
        self : ``Emoji``
            The created emoji.
        """
        emoji_id = cls._last_unicode_id+1
        cls._last_unicode_id = emoji_id
        
        self = object.__new__(cls)
        self.id = emoji_id
        self.animated = False
        self.name = name
        self.unicode = unicode
        self.guild_id = 0
        self.available = True
        self.require_colons = True
        self.managed = False
        self.user = ZEROUSER
        self.role_ids = None
        
        EMOJIS[emoji_id] = self
        UNICODE_TO_EMOJI[unicode] = self
        BUILTIN_EMOJIS[name] = self
        
        for alias in aliases:
            BUILTIN_EMOJIS[alias] = self
        
        return self
    
    
    @property
    def guild(self):
        """
        Returns the guild of the emoji.
        
        Returns `None` if the emoji has no bound emoji, or if the emoji's guild is not cached.
        
        Returns
        -------
        guild : `None` or ``Guild``
            The emoji's guild.
        """
        guild_id = self.guild_id
        if guild_id:
            return GUILDS.get(guild_id, None)
    
    
    @property
    def roles(self):
        """
        Returns the roles for which's members the emoji's usage is restricted to.
        
        Returns
        -------
        roles : `None` or `tuple` of ``Role``
        """
        role_ids = self.role_ids
        if role_ids is None:
            roles = None
        else:
            roles = tuple(sorted(
                (create_partial_role_from_id(role_id) for role_id in role_ids),
                key = id_sort_key,
            ))
        
        return roles



import sys
from os.path import dirname as get_directory_name, realpath as get_real_path, join as join_paths, \
    normpath as normalize_path, expanduser as get_user_home_directory
from os import getcwd as get_current_work_directory

try:
    from . import __package__ as PACKAGE_NAME
except ImportError:
    # If we have hata not setuped
    PACKAGE_NAME = sys.path[0]
    
    sys.path.append(
        normalize_path(
            join_paths(
                get_directory_name(
                    get_real_path(
                        join_paths(
                            get_current_work_directory(),
                            get_user_home_directory(__file__),
                        )
                    )
                ),
                '..',
            )
        )
    )


PACKAGE = __import__(PACKAGE_NAME)

SYSTEM_DEFAULT_PARAMETER = 'i'

COMMAND_NAMES = tuple(sorted((
    'help',
    'interpreter',
    'version',
)))

COMMAND_MAP = {
    'h': 'help',
    'help': 'help',
    
    'i': 'interpreter',
    'interpreter': 'interpreter',
    
    'v': 'version',
    'version': 'version',
}

assert SYSTEM_DEFAULT_PARAMETER in COMMAND_MAP


def command_not_found():
    from hata.backend.utils import get_short_executable
    
    output_parts = ['No command is added for: ']
    
    system_parameter = sys.argv
    
    index = 1
    length = len(system_parameter)
    
    while True:
        output_parts.append(repr(system_parameter[index]))
        index += 1
        if index == length:
            break
        
        output_parts.append(', ')
        continue
    
    output_parts.append('\n')
    output_parts.append('Try using "$')
    output_parts.append(get_short_executable())
    output_parts.append(' ')
    output_parts.append(PACKAGE_NAME)
    output_parts.append(' help" for more information\n.')
    
    output = ''.join(output_parts)
    
    sys.stderr.write(output)


def __main__():
    system_parameters = sys.argv
    if len(system_parameters) < 2:
        system_parameter = SYSTEM_DEFAULT_PARAMETER
    else:
        system_parameter = system_parameters[1].lower()
    
    try:
        command_name = COMMAND_MAP[system_parameter]
    except KeyError:
        return command_not_found
    
    __import__(f'{PACKAGE_NAME}.main.{command_name}')
    return getattr(PACKAGE.main, command_name).__main__


if __name__ == '__main__':
    # Do tail call.
    __main__()()



__all__ = ('require', )

from sys import path as route_paths, _getframe as get_frame
from os.path import join as join_paths, isdir as is_folder, isfile as is_file, exists, basename as base_name, \
    isabs as is_absolute_path_name
from os import listdir as list_directory
from ...backend.utils import HybridValueDictionary
from ...backend.analyzer import CallableAnalyzer
from .exceptions import DoNotLoadExtension

def _validate_entry_or_exit(point):
    """
    Validates the given entry or exit point, returning `True`, if they passed.
    
    Parameters
    ----------
    point : `None`, `str` or `callable`
        The point to validate.
    
    Raises
    ------
    TypeError
        If `point` was given as `callable`, but accepts less or more positional parameters, as would be given.
    """
    if point is None:
        return True
    
    if isinstance(point, str):
        return True
    
    if callable(point):
        analyzer = CallableAnalyzer(point)
        min_, max_ = analyzer.get_non_reserved_positional_parameter_range()
        if min_ > 1:
            raise TypeError(f'`{point!r}` excepts at least `{min_!r}` non reserved parameters, meanwhile the event '
                'expects to pass `1`.')
        
        if min_ == 1:
            return True
        
        #min<expected
        if max_ >= 1:
            return True
        
        if analyzer.accepts_args():
            return True
        
        raise TypeError(f'`{point!r}` expects maximum `{max_!r}` non reserved parameters  meanwhile the event expects '
            'to pass `1`.')
    
    return False


def validate_extension_parameters(entry_point=None, exit_point=None, extend_default_variables=True, locked=False,
        take_snapshot_difference=True, **variables):
    """
    Validates extension parameters.
    
    Parameters
    ----------
    entry_point : `None`, `str` or `callable`, Optional
        Extension specific entry point, to use over the extension loader's default.
    exit_point : `None`, `str` or `callable`, Optional
        Extension specific exit point, to use over the extension loader's default.
    locked : `bool`, Optional
        Whether the given extension(s) should not be affected by `.{}_all` methods.
    take_snapshot_difference : `bool`, Optional
        Whether snapshot feature should be used.
    **variables : Keyword parameters
        Variables to assign to an extension(s)'s module before they are loaded.
    
    Raises
    ------
    TypeError
        - If `entry_point` was not given as `None`, `str` or as `callable`.
        - If `entry_point` was given as `callable`, but accepts less or more positional parameters, as would be
            given.
        - If `exit_point` was not given as `None`, `str` or as `callable`.
        - If `exit_point` was given as `callable`, but accepts less or more positional parameters, as would be
            given.
        - If `extend_default_variables` was not given as `bool`.
        - If `locked` was not given as `bool`.
        - If `name` was not given as `str` or as `iterable` of `str`.
    ValueError
        If a variable name is would be used, what is `module` attribute.
    
    Returns
    -------
    entry_point : `None`, `str` or `callable`
        Extension specific entry point, to use over the extension loader's default.
    exit_point : `None`, `str` or `callable`
        Extension specific exit point, to use over the extension loader's default.
    extend_default_variables : `bool`
        Whether the extension should use the loader's default variables or just it's own's.
    locked : `bool`
        Whether the given extension(s) should not be affected by `.{}_all` methods.
    take_snapshot_difference : `bool`
        Whether snapshot feature should be used.
    default_variables : `None` or `HybridValueDictionary` of (`str`, `Any`) items
        An optionally weak value dictionary to store objects for assigning them to modules before loading them.
        If would be empty, is set as `None` instead.
    """
    if not _validate_entry_or_exit(entry_point):
        raise TypeError(f'`validate_extension_parameters` expected `None`, `str` or a `callable` as '
            f'`entry_point`, got {entry_point.__class__.__name__}.')
    
    if not _validate_entry_or_exit(exit_point):
        raise TypeError(f'`validate_extension_parameters` expected `None`, `str` or a `callable` as `exit_point`, '
            f'got {exit_point.__class__.__name__}.')
    
    if variables:
        default_variables = HybridValueDictionary(variables)
        for key, value in variables.items():
            if key in PROTECTED_NAMES:
                raise ValueError(f'The passed {key!r} is a protected variable name of module type.')
            default_variables[key] = value
    else:
        default_variables = None
    
    extend_default_variables_type = extend_default_variables.__class__
    if extend_default_variables_type is bool:
        pass
    elif issubclass(extend_default_variables_type, int):
        extend_default_variables = bool(extend_default_variables)
    else:
        raise TypeError(f'`extend_default_variables` should have been passed as `bool`, got: '
            f'{extend_default_variables_type.__name__}.')
    
    locked_type = type(locked)
    if locked_type is bool:
        pass
    elif issubclass(locked_type, int):
        locked = bool(locked)
    else:
        raise TypeError(f'`locked` should have been passed as `bool`, got: {locked_type.__name__}.')
    
    return entry_point, exit_point, extend_default_variables, locked, take_snapshot_difference, default_variables


PROTECTED_NAMES = {
    '__class__',
    '__delattr__',
    '__dict__',
    '__dir__',
    '__doc__',
    '__eq__',
    '__format__',
    '__ge__',
    '__getattribute__',
    '__gt__',
    '__hash__',
    '__init__',
    '__init_subclass__',
    '__le__',
    '__lt__',
    '__module__',
    '__ne__',
    '__new__',
    '__reduce__',
    '__reduce_ex__',
    '__repr__',
    '__setattr__',
    '__sizeof__',
    '__str__',
    '__subclasshook__',
    '__weakref__',
    '_cached',
    '_set_fileattr',
    'cached',
    'has_location',
    'loader',
    'loader_state',
    'name',
    'origin',
    'parent',
    'submodule_search_locations',
}



PYTHON_EXTENSION_NAMES = (
    '.py',
    '.pyd',
    '.pyc',
    '.so',
)


def _iter_extension_names_and_paths(name):
    """
    Fetches the names
    
    This function is a generator.
    
    Parameters
    ----------
    name : `str` or `iterable` of `str`
        The name to fetch to single strings.
    
    Yields
    ------
    name : `str`
        Extension names.
    path : `str`
        Path of the extension file.
    
    Raises
    ------
    ImportError
        If a name could not be detected as an extension.
    TypeError
        If `name` is not `str` nor an `iterable` of `str` instances.
    """
    for name in _iter_name_maybe_iterable(name):
        if name.startswith(ABSOLUTE_PATH_EXTENSION_NAME_PREFIX):
            return name
        
        yield from _lookup_path(name)


def _iter_name_maybe_iterable(name):
    """
    Fetches the given name.
    
    This function is a generator.
    
    Parameters
    ----------
    name : `str` or `iterable` of `str`
        The name to fetch to single strings.
    
    Yields
    ------
    name : `str`
        Extension names.
    
    Raises
    ------
    TypeError
        If `name` is not `str` nor an `iterable` of `str` instances.
    """
    name_type = type(name)
    if name_type is str:
        yield name
    elif issubclass(name_type, str):
        yield str(name)
    elif hasattr(name_type, '__iter__'):
        for sub_name in name:
            sub_name_type = type(sub_name)
            if sub_name_type is str:
                yield sub_name
            elif issubclass(sub_name_type, str):
                yield str(sub_name)
            else:
                raise TypeError(f'`name` can be given as an `str` or as `iterable` of `str`, got an `iterable`, which '
                    f'contains an: {sub_name_type.__name__} element.')
    else:
        raise TypeError(f'`name` can be given as an `str` or as `iterable` of `str`, got an `iterable`, got '
            f'{name_type.__name__}.')


def _lookup_path(import_name_or_path):
    """
    Detects the root of the given name.
    
    This function is an iterable generator.
    
    Parameters
    ----------
    import_name_or_path : `str`
        An extension's import name, or it's absolute path.
    
    Yields
    ------
    import_name : `None` or `str`
        Import name to an extension file.
    path : `str`
        Path of the file.
    
    Raise
    -----
    ImportError
        If `import_name_or_path` name could not be detected as en extension.
    """
    if is_absolute_path_name(import_name_or_path):
        if exists(import_name_or_path):
            if is_folder(import_name_or_path):
                yield from _iter_folder(None, import_name_or_path)
                return
            
            if is_file(import_name_or_path):
                yield None, import_name_or_path
                return
    else:
        path_end = join_paths(*import_name_or_path.split('.'))
        for base_path in route_paths:
            path = join_paths(base_path, path_end)
            if exists(path) and is_folder(path):
                yield from _iter_folder(import_name_or_path, path)
                return
            
            for python_extension_name in PYTHON_EXTENSION_NAMES:
                file_path = path+python_extension_name
                if exists(file_path) and is_file(file_path):
                    yield import_name_or_path, file_path
                    return
    
    raise ImportError(f'The given `import_name_or_path` could not be detected as an extension nor an absolute path, '
        f'got {import_name_or_path!r}.')


def _iter_folder(import_name, folder_path):
    """
    Iterates over a folder's import names.
    
    Parameters
    ----------
    import_name : `None` or `str`
        The name of the extension if we would import it.
    folder_path : `str`
        Path to the folder
    
    Yields
    ------
    import_name : `None` or `str`
        Detected import names for each applicable file in the folder.
    path : `str`
        Path of the file.
    """
    for python_extension_name in PYTHON_EXTENSION_NAMES:
        file_path = join_paths(folder_path, f'__init__{python_extension_name}')
        if exists(file_path) and is_file(file_path):
            yield import_name, file_path
            return
    
    for file_name in list_directory(folder_path):
        if file_name.startswith('.') or (file_name == '__pycache__'):
            continue
        
        path = join_paths(folder_path, file_name)
        
        if is_file(path):
            for python_extension_name in PYTHON_EXTENSION_NAMES:
                if file_name.endswith(python_extension_name):
                    if import_name is None:
                        import_name_value = None
                    else:
                        import_name_value = f'{import_name}.{file_name[:-len(python_extension_name)]}'
                    yield import_name_value, path
                    break
            
            continue
        
        if is_folder(path):
            if import_name is None:
                import_name_value = None
            else:
                import_name_value = f'{import_name}.{file_name}'
            yield from _iter_folder(import_name_value, path)
            continue
        
        # no more cases
        continue



def require(*args, **kwargs):
    """
    Requires the given parameters.
    
    Parameters
    ----------
    *args : Parameters
        Required variable names.
    **kwargs : Keyword parameters
        Variables and their expected value / type.
    """
    module_globals = get_frame().f_back.f_globals
    
    for variable_name in args:
        if variable_name not in module_globals:
            raise DoNotLoadExtension(variable_name)
    
    for variable_name, expected_value in kwargs.items():
        try:
            variable_value = module_globals[variable_name]
        except KeyError:
            raise DoNotLoadExtension(variable_name) from None
        
        if variable_value is expected_value:
            continue
        
        if isinstance(expected_value, type) and isinstance(variable_value, expected_value):
            continue
        
        raise DoNotLoadExtension(variable_name, variable_value, expected_value)


ABSOLUTE_PATH_EXTENSION_NAME_PREFIX = '<extension>.'

def _get_path_extension_name(path):
    """
    Creates extension name from the given path.
    
    Parameter
    ---------
    path : `str`
        Path to a file.
    
    Returns
    -------
    extension_name : `str`
    """
    file_name = base_name(path)
    dot_index = file_name.rfind('.')
    if dot_index != -1:
        file_name = file_name[:dot_index]
    
    return ABSOLUTE_PATH_EXTENSION_NAME_PREFIX+file_name



from .preinstanced import *
from .utils import *
from .webhook import *
from .webhook_base import *
from .webhook_repr import *

__all__ = (
    *preinstanced.__all__,
    *utils.__all__,
    *webhook.__all__,
    *webhook_base.__all__,
    *webhook_repr.__all__,
)



__all__ = ('Stage',)

import reprlib

from ...backend.export import export, include

from ..bases import DiscordEntity
from ..channel import ChannelStage, create_partial_channel_from_id
from ..core import STAGES

from ..scheduled_event import PrivacyLevel

Guild = include('Guild')
create_partial_guild_from_id = include('create_partial_guild_from_id')

@export
class Stage(DiscordEntity):
    """
    Represents an active stage instance of a stage channel.
    
    Attributes
    ----------
    id : `int`
        The stage instance's identifier.
    channel : ``ChannelStage``
        The stage channel where the stage is active.
    discoverable : `bool`
        Whether the stage is discoverable.
    guild : ``Guild``
        The guild of the stage channel.
    invite_code : `None` or `str`
        Invite code to the stage's voice channel.
    privacy_level : ``PrivacyLevel``
        The privacy level of the stage.
    scheduled_event_id : `int`
        Whether the stage was started by a scheduled event.
    topic : `str`
        The topic of the stage. Can be empty string.
    """
    __slots__ = ('__weakref__', 'channel', 'discoverable', 'guild', 'invite_code', 'privacy_level',
        'scheduled_event_id', 'topic')
    
    def __new__(cls, data):
        """
        Creates a new stage instance from the received data.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Stage data.
        """
        stage_id = int(data['id'])
        try:
            self = STAGES[stage_id]
        except KeyError:
            self = object.__new__(cls)
            guild = create_partial_guild_from_id(int(data['guild_id']))
            
            self.channel = create_partial_channel_from_id(int(data['channel_id']), 13, guild.id)
            self.guild = guild
            self.id = stage_id
            
            scheduled_event_id = data.get('guild_scheduled_event_id', None)
            if scheduled_event_id is None:
                scheduled_event_id = 0
            else:
                scheduled_event_id = int(scheduled_event_id)
            self.scheduled_event_id = scheduled_event_id
            
            self._update_attributes(data)
            
            stages = guild.stages
            if stages is None:
                stages = guild.stages = {}
            
            stages[stage_id] = self
            
            STAGES[stage_id] = self
        
        return self
    
    
    def __repr__(self):
        """Returns the stage's representation."""
        repr_parts = [
            '<',
            self.__class__.__name__,
            ' channel=',
            repr(self.channel),
            ', topic=',
            reprlib.repr(self.topic),
            '>',
        ]
        
        return ''.join(repr_parts)
    
    
    def _update_attributes(self, data):
        """
        Updates the stage from the given data.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Stage data.
        """
        self.topic = data['topic']
        self.invite_code = data.get('invite_code', None)
        self.discoverable = not data.get('discoverable_disabled', False)
        self.privacy_level = PrivacyLevel.get(data.get('privacy_level', 2))
    
    
    def _difference_update_attributes(self, data):
        """
        Updates the stage from the given data and returns the changed attributes in `attribute-name` - `old-value`
        relation.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Stage data.
        
        Returns
        -------
        old_attributes : `dict` of (`str`, `Any`) items
            The changed attributes of the stage.
            
            Each item in the returned dictionary is optional.
        
        Returned Data Structure
        -----------------------
        +---------------+-----------------------+
        | Keys          | Values                |
        +===============+=======================+
        | discoverable  | `bool`                |
        +---------------+-----------------------+
        | invite_code   | `None` or `str`       |
        +---------------+-----------------------+
        | privacy_level | ``PrivacyLevel`` |
        +---------------+-----------------------+
        | topic         | `str`                 |
        +---------------+-----------------------+
        """
        old_attributes = {}
        
        topic = data['topic']
        if topic != self.topic:
            old_attributes['topic'] = self.topic
            self.topic = topic
        
        
        invite_code = data.get('invite_code', None)
        if invite_code != self.invite_code:
            old_attributes['invite_code'] = self.invite_code
            self.invite_code = invite_code
        
        
        discoverable = not data.get('discoverable_disabled', False)
        if discoverable != self.discoverable:
            old_attributes['discoverable'] = self.discoverable
            self.discoverable = discoverable
        
        
        privacy_level = PrivacyLevel.get(data.get('privacy_level', 2))
        if privacy_level is not self.privacy_level:
            privacy_level['privacy_level'] = self.privacy_level
            self.privacy_level = privacy_level
        
        return old_attributes
    
    
    def _delete(self):
        """
        Removes the stage's references.
        """
        guild = self.guild
        stages = guild.stages
        if (stages is not None):
            try:
                del stages[self.id]
            except KeyError:
                pass
            else:
                if not stages:
                    guild.stages = None
        
        try:
            del STAGES[self.id]
        except KeyError:
            pass
